# Generated by Django 3.1.4 on 2021-03-08 15:59

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("users", "0001_squashed_0009_auto_20210219_1237"),
    ]

    operations = [
        migrations.AlterField(
            model_name="user",
            name="activity_at",
            field=models.DateTimeField(auto_now=True, verbose_name="last annotation activity"),
        ),
        migrations.AlterField(
            model_name="user",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="Designates whether to treat this user as active. Unselect this instead of deleting accounts.",
                verbose_name="active",
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""

from .models import Organization, OrganizationMember


from django.forms import (
    Textarea,
    ModelForm,
    TextInput,
    Form,
    FileField,
    Select,
    HiddenInput,
    CharField,
)


class OrganizationForm(ModelForm):
    """ """

    org_update = CharField(widget=HiddenInput(), required=False)

    class Meta:
        model = Organization
        fields = ("title",)


class OrganizationSignupForm(ModelForm):
    """ """

    class Meta:
        model = Organization
        fields = ("title",)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 3.1.4 on 2021-03-03 07:31

import core.utils.common
from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


def rename_disabled_to_off0006(apps, schema_editor):
    OrganizationMember = apps.get_model("organizations", "OrganizationMember")
    OrganizationMember.objects.filter(role="Disabled").update(role="Off")

    migrations.AlterField(
        model_name="organizationmember",
        name="role",
        field=models.CharField(
            choices=[
                ("Administrator", "Administrator"),
                ("Manager", "Manager"),
                ("Coordinator", "Coordinator"),
                ("Collaborator", "Collaborator"),
                ("Not Activated", "Not Activated"),
                ("Off", "Off"),
            ],
            default="Not Activated",
            help_text="Organization membership role",
            max_length=100,
        ),
    )


def rename_disabled_to_off0007(apps, schema_editor):
    OrganizationMember = apps.get_model("organizations", "OrganizationMember")
    OrganizationMember.objects.filter(role="Off").update(role="Deactivated")

    migrations.AlterField(
        model_name="organizationmember",
        name="role",
        field=models.CharField(
            choices=[
                ("Administrator", "Administrator"),
                ("Manager", "Manager"),
                ("Coordinator", "Coordinator"),
                ("Collaborator", "Collaborator"),
                ("Not Activated", "Not Activated"),
                ("Deactivated", "Deactivated"),
            ],
            default="Not Activated",
            help_text="Organization membership role",
            max_length=100,
        ),
    )


class Migration(migrations.Migration):

    replaces = [
        ("organizations", "0001_initial"),
        ("organizations", "0002_organization_token"),
        ("organizations", "0003_auto_20200418_0202"),
        ("organizations", "0004_auto_20200501_1751"),
        ("organizations", "0005_auto_20200811_2313"),
        ("organizations", "0006_auto_20200923_1423"),
        ("organizations", "0007_auto_20200923_2200"),
        ("organizations", "0008_auto_20201005_1552"),
    ]

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("title", models.CharField(max_length=1000, verbose_name="organization title")),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "created_by",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="organization",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created_by",
                    ),
                ),
            ],
            options={
                "db_table": "organization",
            },
        ),
        migrations.CreateModel(
            name="OrganizationMember",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[
                            ("Administrator", "Administrator"),
                            ("Manager", "Manager"),
                            ("Coordinator", "Coordinator"),
                            ("Collaborator", "Collaborator"),
                            ("Not Activated", "Not Activated"),
                            ("Disabled", "Disabled"),
                        ],
                        default="Not Activated",
                        max_length=100,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "organization",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="organizations.organization"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="om_through",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="organization",
            name="users",
            field=models.ManyToManyField(
                related_name="organizations",
                through="organizations.OrganizationMember",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="token",
            field=models.CharField(
                blank=True,
                default=core.utils.common.create_hash,
                max_length=256,
                null=True,
                unique=True,
                verbose_name="token",
            ),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="organization",
            field=models.ForeignKey(
                help_text="Organization ID",
                on_delete=django.db.models.deletion.CASCADE,
                to="organizations.organization",
            ),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="role",
            field=models.CharField(
                choices=[
                    ("Administrator", "Administrator"),
                    ("Manager", "Manager"),
                    ("Coordinator", "Coordinator"),
                    ("Collaborator", "Collaborator"),
                    ("Not Activated", "Not Activated"),
                    ("Disabled", "Disabled"),
                ],
                default="Not Activated",
                help_text="Organization membership role",
                max_length=100,
            ),
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="user",
            field=models.ForeignKey(
                help_text="User ID",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="om_through",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="organization",
            name="default_role",
            field=models.CharField(
                default="Not Activated",
                help_text="Default membership role for invited users",
                max_length=100,
            ),
        ),
        migrations.RunPython(
            code=rename_disabled_to_off0006,
        ),
        migrations.RunPython(
            code=rename_disabled_to_off0007,
        ),
        migrations.AlterField(
            model_name="organizationmember",
            name="role",
            field=models.CharField(
                choices=[
                    ("Administrator", "Administrator"),
                    ("Manager", "Manager"),
                    ("Coordinator", "Coordinator"),
                    ("Collaborator", "Collaborator"),
                    ("Not Activated", "Not Activated"),
                    ("Deactivated", "Deactivated"),
                ],
                default="Not Activated",
                help_text="Organization membership role",
                max_length=100,
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from core.label_config import generate_sample_task_without_check
from drf_dynamic_fields import DynamicFieldsMixin
from rest_framework import serializers
from rest_framework.serializers import SerializerMethodField
from users.serializers import UserSimpleSerializer

from projects.models import Project, ProjectOnboarding, ProjectSummary


class CreatedByFromContext:
    requires_context = True

    def __call__(self, serializer_field):
        return serializer_field.context.get("created_by")


class ProjectSerializer(DynamicFieldsMixin, serializers.ModelSerializer):
    """Serializer get numbers from project queryset annotation,
    make sure, that you use correct one(Project.objects.with_counts())
    """

    task_number = serializers.IntegerField(
        default=None, read_only=True, help_text="Total task number in project"
    )
    total_annotations_number = serializers.IntegerField(
        default=None,
        read_only=True,
        help_text="Total annotations number in project including "
        "skipped_annotations_number and ground_truth_number.",
    )
    total_predictions_number = serializers.IntegerField(
        default=None,
        read_only=True,
        help_text="Total predictions number in project including "
        "skipped_annotations_number and ground_truth_numberuseful_annotation_number.",
    )
    useful_annotation_number = serializers.IntegerField(
        default=None,
        read_only=True,
        help_text="Useful annotation number in project not including "
        "skipped_annotations_number and ground_truth_number. "
        "Total annotations = annotation_number + "
        "skipped_annotations_number + ground_truth_number",
    )
    ground_truth_number = serializers.IntegerField(
        default=None, read_only=True, help_text="Honeypot annotation number in project"
    )
    skipped_annotations_number = serializers.IntegerField(
        default=None,
        read_only=True,
        help_text="Skipped by collaborators annotation number in project",
    )
    created_by = UserSimpleSerializer(default=CreatedByFromContext())

    parsed_label_config = SerializerMethodField(
        default=None, read_only=True, help_text="JSON-formatted labeling configuration"
    )
    start_training_on_annotation_update = SerializerMethodField(
        default=None,
        read_only=False,
        help_text="Start model training after any annotations are submitted or updated",
    )
    config_has_control_tags = SerializerMethodField(
        default=None, read_only=True, help_text="Flag to detect is project ready for labeling"
    )

    @staticmethod
    def get_config_has_control_tags(project):
        return len(project.get_control_tags_from_config()) > 0

    @staticmethod
    def get_parsed_label_config(project):
        return project.get_parsed_config()

    def get_start_training_on_annotation_update(self, instance):
        # FIXME: remake this logic with start_training_on_annotation_update
        return True if instance.min_annotations_to_start_training else False

    def to_internal_value(self, data):
        # FIXME: remake this logic with start_training_on_annotation_update
        initial_data = data
        data = super().to_internal_value(data)
        if "start_training_on_annotation_update" in initial_data:
            data["min_annotations_to_start_training"] = int(
                initial_data["start_training_on_annotation_update"]
            )
        return data

    class Meta:
        model = Project
        extra_kwargs = {
            "memberships": {"required": False},
            "title": {"required": False},
            "created_by": {"required": False},
        }
        fields = [
            "id",
            "title",
            "description",
            "label_config",
            "expert_instruction",
            "show_instruction",
            "show_skip_button",
            "enable_empty_annotation",
            "show_annotation_history",
            "organization",
            "color",
            "maximum_annotations",
            "is_published",
            "model_version",
            "is_draft",
            "created_by",
            "created_at",
            "min_annotations_to_start_training",
            "start_training_on_annotation_update",
            "show_collab_predictions",
            "num_tasks_with_annotations",
            "task_number",
            "useful_annotation_number",
            "ground_truth_number",
            "skipped_annotations_number",
            "total_annotations_number",
            "total_predictions_number",
            "sampling",
            "show_ground_truth_first",
            "show_overlap_first",
            "overlap_cohort_percentage",
            "task_data_login",
            "task_data_password",
            "control_weights",
            "parsed_label_config",
            "evaluate_predictions_automatically",
            "config_has_control_tags",
        ]

    def validate_label_config(self, value):
        if self.instance is None:
            # No project created yet
            Project.validate_label_config(value)
        else:
            # Existing project is updated
            self.instance.validate_config(value)
        return value


class ProjectOnboardingSerializer(serializers.ModelSerializer):
    class Meta:
        model = ProjectOnboarding
        fields = "__all__"


class ProjectLabelConfigSerializer(serializers.Serializer):
    label_config = serializers.CharField()

    def validate_label_config(self, config):
        Project.validate_label_config(config)
        return config


class ProjectSummarySerializer(serializers.ModelSerializer):
    class Meta:
        model = ProjectSummary
        fields = "__all__"



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import json
import logging
import os
import datetime

from urllib.parse import urljoin

from django.conf import settings
from django.db import models, connection
from django.db.models import Q, F, When, Count, Case, Subquery, OuterRef, Value
from django.db.models.functions import Coalesce
from django.db.models.signals import post_delete, pre_save, post_save, pre_delete
from django.db.utils import ProgrammingError, OperationalError, DatabaseError
from django.utils.translation import gettext_lazy as _
from django.db.models import JSONField
from django.urls import reverse
from django.utils.timesince import timesince
from django.utils.timezone import now
from django.dispatch import receiver, Signal

from model_utils import FieldTracker

from core.utils.common import (
    find_first_one_to_one_related_field_by_prefix,
    string_is_url,
    load_func,
)
from core.utils.params import get_env
from data_manager.managers import PreparedTaskManager, TaskManager

logger = logging.getLogger(__name__)

TaskMixin = load_func(settings.TASK_MIXIN)


class Task(TaskMixin, models.Model):
    """Business tasks from project"""

    id = models.AutoField(
        auto_created=True, primary_key=True, serialize=False, verbose_name="ID", db_index=True
    )
    data = JSONField(
        "data",
        null=False,
        help_text="User imported or uploaded data for a task. Data is formatted according to "
        "the project label config. You can find examples of data for your project "
        "on the Import page in the Label Studio Data Manager UI.",
    )
    meta = JSONField(
        "meta",
        null=True,
        default=dict,
        help_text="Meta is user imported (uploaded) data and can be useful as input for an ML "
        "Backend for embeddings, advanced vectors, and other info. It is passed to "
        "ML during training/predicting steps.",
    )
    project = models.ForeignKey(
        "projects.Project",
        related_name="tasks",
        on_delete=models.CASCADE,
        null=True,
        help_text="Project ID for this task",
    )
    created_at = models.DateTimeField(
        _("created at"), auto_now_add=True, help_text="Time a task was created"
    )
    updated_at = models.DateTimeField(
        _("updated at"), auto_now=True, help_text="Last time a task was updated"
    )
    is_labeled = models.BooleanField(
        _("is_labeled"),
        default=False,
        help_text="True if the number of annotations for this task is greater than or equal "
        "to the number of maximum_completions for the project",
    )
    overlap = models.IntegerField(
        _("overlap"),
        default=1,
        db_index=True,
        help_text="Number of distinct annotators that processed the current task",
    )
    file_upload = models.ForeignKey(
        "data_import.FileUpload",
        on_delete=models.SET_NULL,
        null=True,
        blank=True,
        related_name="tasks",
        help_text="Uploaded file used as data source for this task",
    )
    updates = ["is_labeled"]

    objects = TaskManager()  # task manager by default
    prepared = (
        PreparedTaskManager()
    )  # task manager with filters, ordering, etc for data_manager app

    @property
    def file_upload_name(self):
        return os.path.basename(self.file_upload.file.name)

    @classmethod
    def get_locked_by(cls, user, project=None, tasks=None):
        """Retrieve the task locked by specified user. Returns None if the specified user didn't lock anything."""
        if project:
            lock = TaskLock.objects.filter(
                user=user, expire_at__gt=now(), task__project=project
            ).first()
        elif tasks:
            return tasks.filter(locks__user=user, locks__expire_at__gt=now()).first()
        else:
            raise Exception("Neither project or tasks passed to get_locked_by")

        if lock:
            return lock.task

    def has_lock(self):
        """Check whether current task has been locked by some user"""
        num_locks = self.num_locks
        num_annotations = self.annotations.filter(ground_truth=False).count()
        num = num_locks + num_annotations
        if num > self.overlap:
            logger.error(
                f"Num takes={num} > overlap={self.overlap} for task={self.id} - it's a bug"
            )
        return num >= self.overlap

    @property
    def num_locks(self):
        return self.locks.filter(expire_at__gt=now()).count()

    def get_lock_ttl(self):
        if settings.TASK_LOCK_TTL is not None:
            return settings.TASK_LOCK_TTL
        avg_lead_time = self.project.annotations_lead_time()
        return (
            3 * int(avg_lead_time) if avg_lead_time is not None else settings.TASK_LOCK_DEFAULT_TTL
        )

    def has_permission(self, user):
        return self.project.has_permission(user)

    def clear_expired_locks(self):
        self.locks.filter(expire_at__lt=now()).delete()

    def set_lock(self, user):
        """Lock current task by specified user. Lock lifetime is set by `expire_in_secs`"""
        num_locks = self.num_locks
        if num_locks < self.overlap:
            expire_at = now() + datetime.timedelta(seconds=self.get_lock_ttl())
            TaskLock.objects.create(task=self, user=user, expire_at=expire_at)
            logger.debug(f"User={user} acquires a lock for the task={self}")
        else:
            logger.error(
                f"Current number of locks for task {self.id} is {num_locks}, but overlap={self.overlap}: "
                f"that's a bug because this task should not be taken in a label stream (task should be locked)"
            )
        self.clear_expired_locks()

    def release_lock(self, user=None):
        """Release lock for the task.
        If user specified, it checks whether lock is released by the user who previously has locked that task"""

        if user is not None:
            self.locks.filter(user=user).delete()
        else:
            self.locks.all().delete()
        self.clear_expired_locks()

    def get_storage_link(self):
        # TODO: how to get neatly any storage class here?
        return find_first_one_to_one_related_field_by_prefix(self, "io_storages_")

    def resolve_uri(self, task_data, proxy=True):
        if proxy and self.project.task_data_login and self.project.task_data_password:
            protected_data = {}
            for key, value in task_data.items():
                if isinstance(value, str) and string_is_url(value):
                    path = (
                        reverse("projects-file-proxy", kwargs={"pk": self.project.pk})
                        + "?url="
                        + value
                    )
                    value = urljoin(settings.HOSTNAME, path)
                protected_data[key] = value
            return protected_data
        else:
            # Try resolve URLs via storage associated with that task
            storage = self._get_task_storage()
            if storage:
                return storage.resolve_task_data_uri(task_data)
            return task_data

    def _get_task_storage(self):
        # maybe task has storage link
        storage_link = self.get_storage_link()
        if storage_link:
            return storage_link.storage

        # or try global storage settings (only s3 for now)
        elif get_env("USE_DEFAULT_S3_STORAGE", default=False, is_bool=True):
            # TODO: this is used to access global environment storage settings.
            # We may use more than one and non-default S3 storage (like GCS, Azure)
            from io_storages.s3.models import S3ImportStorage

            return S3ImportStorage()

    def update_is_labeled(self):
        """Set is_labeled field according to annotations*.count > overlap"""
        n = self.annotations.filter(Q_finished_annotations & Q(ground_truth=False)).count()
        # self.is_labeled = n >= self.project.maximum_annotations
        self.is_labeled = n >= self.overlap

    def reset_updates(self):
        """Reset updates to default from model for one task.
        We need it in duplicate project or total deletion of annotations
        """
        for field in Task._meta.fields:
            if field.name in Task.updates:
                setattr(self, field.name, field.default)

    @staticmethod
    def bulk_reset_updates(project):
        """Bulk reset updates to default, it's a fast way to reset all tasks in project"""
        for field in Task._meta.fields:
            if field.name in Task.updates:
                project.tasks.update(**{field.name: field.default})

    @staticmethod
    def bulk_update_is_labeled(project):
        """Fast way to update only is_labeled.
        Prefer to use Django 2.2 bulk_update(), see bulk_update_field('is_labeled')

        get all project.tasks as subquery
        Subquery(
            w coalesce get the first non-null value (count(annotations), or 0)
            make condition
            add temp field pre_is_labeled as condtion values
        )
        update all tasks with Subquery
        """
        tasks = project.tasks.filter(pk=OuterRef("pk"))
        count = Coalesce(
            Count(
                "annotations",
                filter=Q(annotations__was_cancelled=False) & Q(annotations__ground_truth=False),
            ),
            Value(0),
        )
        condition = Case(
            When(overlap__lte=count, then=Value(True)),
            default=Value(False),
            output_field=models.BooleanField(null=False),
        )
        results = tasks.annotate(pre_is_labeled=condition).values("pre_is_labeled")
        project.tasks.update(is_labeled=Subquery(results))

    def delete_url(self):
        return reverse("tasks:task-delete", kwargs={"pk": self.pk})

    def completion_for_ground_truth(self):
        """1 Get ground_truth completion if task has it, else
        2 Get first completion created by owner of project,
        3 Or the first of somebody if no owner's items.
        It's used for ground_truth selection right on data manager page
        """
        if not self.annotations.exists():
            return None

        # ground_truth already exist
        ground_truth_annotations = self.annotations.filter(ground_truth=True)
        if ground_truth_annotations.exists():
            return ground_truth_annotations.first()

        # owner annotation
        owner_annotations = self.annotations.filter(completed_by=self.project.created_by)
        if owner_annotations.count() > 0:
            return owner_annotations.first()

        # annotator annotation
        return self.annotations.first()

    def increase_project_summary_counters(self):
        if hasattr(self.project, "summary"):
            summary = self.project.summary
            summary.update_data_columns([self])

    def decrease_project_summary_counters(self):
        if hasattr(self.project, "summary"):
            summary = self.project.summary
            summary.remove_data_columns([self])

    class Meta:
        db_table = "task"
        ordering = ["-updated_at"]
        indexes = [
            models.Index(fields=["project", "is_labeled"]),
            models.Index(fields=["id", "overlap"]),
        ]


pre_bulk_create = Signal(providing_args=["objs", "batch_size"])
post_bulk_create = Signal(providing_args=["objs", "batch_size"])


class AnnotationManager(models.Manager):
    def for_user(self, user):
        return self.filter(task__project__organization=user.active_organization)

    def bulk_create(self, objs, batch_size=None):
        pre_bulk_create.send(sender=self.model, objs=objs, batch_size=batch_size)
        res = super(AnnotationManager, self).bulk_create(objs, batch_size)
        post_bulk_create.send(sender=self.model, objs=objs, batch_size=batch_size)
        return res


GET_UNIQUE_IDS = """
with tt as (
    select jsonb_array_elements(tch.result) as item from task_completion_history tch
    where task=%(t_id)s and task_annotation=%(tc_id)s
) select count( distinct tt.item -> 'id') from tt"""

AnnotationMixin = load_func(settings.ANNOTATION_MIXIN)


class Annotation(AnnotationMixin, models.Model):
    """Annotations & Labeling results"""

    objects = AnnotationManager()
    tracker = FieldTracker(fields=["ground_truth", "result"])

    result = JSONField(
        "result",
        null=True,
        default=None,
        help_text="The main value of annotator work - " "labeling result in JSON format",
    )

    task = models.ForeignKey(
        "tasks.Task",
        on_delete=models.CASCADE,
        related_name="annotations",
        null=True,
        help_text="Corresponding task for this annotation",
    )
    completed_by = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="annotations",
        on_delete=models.SET_NULL,
        null=True,
        help_text="User ID of the person who created this annotation",
    )
    was_cancelled = models.BooleanField(
        _("was cancelled"), default=False, help_text="User skipped the task"
    )
    ground_truth = models.BooleanField(
        _("ground_truth"),
        default=False,
        help_text="This annotation is a Ground Truth (ground_truth)",
    )
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")
    updated_at = models.DateTimeField(_("updated at"), auto_now=True, help_text="Last updated time")
    lead_time = models.FloatField(
        _("lead time"),
        null=True,
        default=None,
        help_text="How much time it took to annotate the task",
    )
    prediction = JSONField(
        _("prediction"),
        null=True,
        default=dict,
        help_text="Prediction viewed at the time of annotation",
    )
    result_count = models.IntegerField(
        _("result count"), default=0, help_text="Results inside of annotation counter"
    )

    class Meta:
        db_table = "task_completion"
        indexes = [models.Index(fields=["task", "ground_truth"])]

    def created_ago(self):
        """Humanize date"""
        return timesince(self.created_at)

    def entities_num(self):
        res = self.result
        if isinstance(res, str):
            res = json.loads(res)
        if res is None:
            res = []

        return len(res)

    def has_permission(self, user):
        return self.task.project.has_permission(user)

    def increase_project_summary_counters(self):
        if hasattr(self.task.project, "summary"):
            logger.debug(f"Increase project.summary counters from {self}")
            summary = self.task.project.summary
            summary.update_created_annotations_and_labels([self])

    def decrease_project_summary_counters(self):
        if hasattr(self.task.project, "summary"):
            logger.debug(f"Decrease project.summary counters from {self}")
            summary = self.task.project.summary
            summary.remove_created_annotations_and_labels([self])


class TaskLock(models.Model):
    task = models.ForeignKey(
        "tasks.Task", on_delete=models.CASCADE, related_name="locks", help_text="Locked task"
    )
    expire_at = models.DateTimeField(_("expire_at"))
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="task_locks",
        on_delete=models.CASCADE,
        help_text="User who locked this task",
    )


class AnnotationDraft(models.Model):
    result = JSONField(_("result"), help_text="Draft result in JSON format")
    lead_time = models.FloatField(
        _("lead time"), help_text="How much time it took to annotate the task"
    )
    task = models.ForeignKey(
        "tasks.Task",
        on_delete=models.CASCADE,
        related_name="drafts",
        blank=True,
        null=True,
        help_text="Corresponding task for this draft",
    )
    annotation = models.ForeignKey(
        "tasks.Annotation",
        on_delete=models.CASCADE,
        related_name="drafts",
        blank=True,
        null=True,
        help_text="Corresponding annotation for this draft",
    )
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        related_name="drafts",
        on_delete=models.CASCADE,
        help_text="User who created this draft",
    )

    created_at = models.DateTimeField(_("created at"), auto_now_add=True, help_text="Creation time")
    updated_at = models.DateTimeField(_("updated at"), auto_now=True, help_text="Last update time")

    def created_ago(self):
        """Humanize date"""
        return timesince(self.created_at)

    def has_permission(self, user):
        return self.task.project.has_permission(user)


class Prediction(models.Model):
    """ML backend predictions"""

    result = JSONField("result", null=True, default=dict, help_text="Prediction result")
    score = models.FloatField(_("score"), default=None, help_text="Prediction score", null=True)
    model_version = models.TextField(_("model version"), default="", blank=True, null=True)
    cluster = models.IntegerField(
        _("cluster"), default=None, help_text="Cluster for the current prediction", null=True
    )
    neighbors = JSONField(
        "neighbors", null=True, blank=True, help_text="Array of task IDs of the closest neighbors"
    )
    mislabeling = models.FloatField(
        _("mislabeling"), default=0.0, help_text="Related task mislabeling score"
    )

    task = models.ForeignKey("tasks.Task", on_delete=models.CASCADE, related_name="predictions")
    created_at = models.DateTimeField(_("created at"), auto_now_add=True)
    updated_at = models.DateTimeField(_("updated at"), auto_now=True)

    def created_ago(self):
        """Humanize date"""
        return timesince(self.created_at)

    class Meta:
        db_table = "prediction"


@receiver(post_delete, sender=Task)
def update_all_task_states_after_deleting_task(sender, instance, **kwargs):
    """after deleting_task
    use update_tasks_states for all project
    but call only tasks_number_changed section
    """
    try:
        instance.project.update_tasks_states(
            maximum_annotations_changed=False,
            overlap_cohort_percentage_changed=False,
            tasks_number_changed=True,
        )
    except Exception as exc:
        logger.error("Error in update_all_task_states_after_deleting_task: " + str(exc))


@receiver(pre_delete, sender=Task)
def release_task_lock_before_delete(sender, instance, **kwargs):
    if instance is not None:
        instance.release_lock()


# =========== PROJECT SUMMARY UPDATES ===========


@receiver(pre_delete, sender=Task)
def remove_data_columns(sender, instance, **kwargs):
    """Reduce data column counters afer removing task"""
    instance.decrease_project_summary_counters()


def _task_data_is_not_updated(update_fields):
    if update_fields and list(update_fields) == ["is_labeled"]:
        return True


@receiver(pre_save, sender=Task)
def delete_project_summary_data_columns_before_updating_task(
    sender, instance, update_fields, **kwargs
):
    """Before updating task fields - ensure previous info removed from project.summary"""
    if _task_data_is_not_updated(update_fields):
        # we don't need to update counters when other than task.data fields are updated
        return
    try:
        old_task = sender.objects.get(id=instance.id)
    except Task.DoesNotExist:
        # task just created - do nothing
        return
    old_task.decrease_project_summary_counters()


@receiver(post_save, sender=Task)
def update_project_summary_data_columns(sender, instance, created, update_fields, **kwargs):
    """Update task counters in project summary in case when new task has been created"""
    if _task_data_is_not_updated(update_fields):
        # we don't need to update counters when other than task.data fields are updated
        return
    instance.increase_project_summary_counters()


@receiver(pre_save, sender=Annotation)
def delete_project_summary_annotations_before_updating_annotation(sender, instance, **kwargs):
    """Before updating annotation fields - ensure previous info removed from project.summary"""
    try:
        old_annotation = sender.objects.get(id=instance.id)
    except Annotation.DoesNotExist:
        # annotation just created - do nothing
        return
    old_annotation.decrease_project_summary_counters()


@receiver(post_save, sender=Annotation)
def update_project_summary_annotations_and_is_labeled(sender, instance, created, **kwargs):
    """Update annotation counters in project summary"""
    instance.increase_project_summary_counters()

    if created:
        # If new annotation created, update task.is_labeled state
        logger.debug(f"Update task stats for task={instance.task}")
        instance.task.update_is_labeled()
        instance.task.save(update_fields=["is_labeled"])


@receiver(pre_delete, sender=Annotation)
def remove_project_summary_annotations(sender, instance, **kwargs):
    """Remove annotation counters in project summary followed by deleting an annotation"""
    instance.decrease_project_summary_counters()


# =========== END OF PROJECT SUMMARY UPDATES ===========


def _task_exists_in_db(task):
    try:
        Task.objects.get(id=task.id)
    except Task.DoesNotExist:
        return False
    return True


@receiver(post_delete, sender=Annotation)
def update_is_labeled_after_removing_annotation(sender, instance, **kwargs):
    # Update task.is_labeled state
    task = instance.task
    if _task_exists_in_db(
        task
    ):  # To prevent django.db.utils.DatabaseError: Save with update_fields did not affect any rows.
        logger.debug(f"Update task stats for task={task}")
        instance.task.update_is_labeled()
        instance.task.save(update_fields=["is_labeled"])


@receiver(post_save, sender=Annotation)
def delete_draft(sender, instance, **kwargs):
    task = instance.task
    query_args = {"task": instance.task, "annotation": instance}
    if instance.completed_by is not None:
        query_args["user"] = instance.completed_by
    drafts = AnnotationDraft.objects.filter(**query_args)
    num_drafts = drafts.count()
    drafts.delete()
    logger.debug(f"{num_drafts} drafts removed from task {task} after saving annotation {instance}")


@receiver(post_save, sender=Annotation)
def update_ml_backend(sender, instance, **kwargs):
    if instance.ground_truth:
        return

    project = instance.task.project

    if hasattr(project, "ml_backends") and project.min_annotations_to_start_training:
        annotation_count = Annotation.objects.filter(task__project=project).count()

        # start training every N annotation
        if annotation_count % project.min_annotations_to_start_training == 0:
            for ml_backend in project.ml_backends.all():
                ml_backend.train()


Q_finished_annotations = Q(was_cancelled=False) & Q(result__isnull=False)
Q_task_finished_annotations = Q(annotations__was_cancelled=False) & Q(
    annotations__result__isnull=False
)



from __future__ import print_function

"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
""" Version Lib
Copyright (C) 2019 Maxim Tkachenko 

This library automatically generates version of package based on git.
If 'git desc' is successful it will write version to __version__.py:git_version.  
If 'git desc' is fail it will read __version__.py:git_version.

ATTENTION: do not include version_.py to git! It will affect git commit always!
"""
from subprocess import STDOUT, CalledProcessError, check_output as run
import os
import sys
import logging
import json

VERSION_FILE = "version_.py"


def _write_py(info):
    # go to current dir to package __init__.py
    cwd = os.getcwd()
    d = os.path.dirname(__file__)
    d = d if d else "."
    os.chdir(d)

    info_str = json.dumps(info)

    # write txt
    with open(VERSION_FILE, "w") as f:
        os.chdir(cwd)  # back current dir
        f.write(
            "info = %s\n" % info_str + "\n# This file is automatically generated by version.py"
            "\n# Do not include it to git!\n"
        )


def _read_py():
    # go to current dir to package __init__.py
    cwd = os.getcwd()
    d = os.path.dirname(__file__)
    d = d if d else "."
    sys.path.append(d)
    os.chdir(d)

    # read version
    try:
        version_module = __import__(os.path.splitext(VERSION_FILE)[0])
        return version_module.info
    except ImportError as e:
        logging.warning("Can't read version file: " + VERSION_FILE)
        logging.warning(e)
        return {}
    finally:
        os.chdir(cwd)  # back current dir


# get commit info: message, date, hash, branch
def get_git_commit_info(skip_os=True):

    cwd = os.getcwd()
    d = os.path.dirname(__file__)
    d = d if d else "."
    os.chdir(d)

    try:
        # take version from git
        try:
            desc = run(
                "git describe --long --tags --always --dirty", stderr=STDOUT, shell=True
            ).decode("utf-8")
            info = {
                "message": run("git show -s --format=%s", stderr=STDOUT, shell=True)
                .strip()
                .decode("utf8"),
                "commit": run("git show -s --format=%H", stderr=STDOUT, shell=True)
                .strip()
                .decode("utf8"),
                "date": run("git show -s --format=%ai", stderr=STDOUT, shell=True)
                .strip()
                .decode("utf8"),
                "branch": run("git rev-parse --abbrev-ref HEAD", stderr=STDOUT, shell=True)
                .strip()
                .decode("utf8"),
            }
        except CalledProcessError:
            return _read_py()

        # create package version
        version = desc.lstrip("v").rstrip().replace("-", "+", 1).replace("-", ".")
        # take OS name
        if not skip_os:
            keys = ("ID=", "VERSION_ID=", "RELEASE=")
            with open("/etc/os-release") as f:
                os_version = "".join(
                    str(s).split("=", 1)[1].rstrip().strip('"').replace(".", "")
                    for s in f
                    if str(s).startswith(keys)
                )
                version += "." + os_version
        info["version"] = version

        _write_py(info)
        return info

    except Exception as e:
        raise e

    finally:
        os.chdir(cwd)  # back current dir


def get_git_version(skip_os=True):
    info = get_git_commit_info(skip_os)
    return info.get("version", "")


# get only tag from git
def get_short_version():
    version = get_git_version()
    return version.split("+")[0]


if __name__ == "__main__":
    # init version_.py file
    get_git_version()



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import drf_yasg.openapi as openapi

from django_filters.rest_framework import DjangoFilterBackend
from drf_yasg.utils import swagger_auto_schema

from rest_framework.views import APIView
from rest_framework.response import Response
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.authtoken.models import Token
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework import generics
from rest_framework.permissions import IsAuthenticated

from core.permissions import CanModifyUserOrReadOnly
from users.models import User
from users.serializers import UserSerializer
from users.functions import check_avatar


logger = logging.getLogger(__name__)


class UserAPI(viewsets.ModelViewSet):
    queryset = User.objects.all()
    serializer_class = UserSerializer
    permission_classes = [CanModifyUserOrReadOnly]
    filter_backends = [DjangoFilterBackend]
    parser_classes = (JSONParser, MultiPartParser, FormParser)

    @swagger_auto_schema(auto_schema=None, methods=["delete", "post"])
    @action(detail=True, methods=["delete", "post"])
    def avatar(self, request, pk):
        if request.method == "POST":
            avatar = check_avatar(request.FILES)
            request.user.avatar = avatar
            request.user.save()
            return Response({"detail": "avatar saved"}, status=200)

        elif request.method == "DELETE":
            request.user.avatar = None
            request.user.save()
            return Response(status=204)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Save user details",
        operation_description="Save details for a specific user, such as their name, contact information, "
        "or organization in Label Studio.",
        request_body=UserSerializer,
    )
    def update(self, request, *args, **kwargs):
        return super(UserAPI, self).update(request, *args, **kwargs)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="List users",
        operation_description="List the users that exist on the Label Studio server.",
    )
    def list(self, request, *args, **kwargs):
        return super(UserAPI, self).list(request, *args, **kwargs)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Create new user",
        operation_description="Create a new user for Label Studio.",
        request_body=UserSerializer,
    )
    def create(self, request, *args, **kwargs):
        return super(UserAPI, self).create(request, *args, **kwargs)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Get user info",
        operation_description="Get info about a specific Label Studio user.",
    )
    def retrieve(self, request, *args, **kwargs):
        return super(UserAPI, self).retrieve(request, *args, **kwargs)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Update user details",
        operation_description="Update details for a specific user, such as their name, contact information, "
        "or organization in Label Studio.",
        request_body=UserSerializer,
    )
    def partial_update(self, request, *args, **kwargs):
        return super(UserAPI, self).partial_update(request, *args, **kwargs)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Delete user",
        operation_description="Delete a specific Label Studio user.",
    )
    def destroy(self, request, *args, **kwargs):
        return super(UserAPI, self).destroy(request, *args, **kwargs)


class UserResetTokenAPI(APIView):
    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = User.objects.all()
    permission_classes = (IsAuthenticated,)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Reset user token",
        operation_description="Reset a specific user token.",
        responses={
            201: openapi.Response(
                description="User token response",
                schema=openapi.Schema(
                    description="User token",
                    type=openapi.TYPE_OBJECT,
                    properties={
                        "token": openapi.Schema(description="Token", type=openapi.TYPE_STRING)
                    },
                ),
            )
        },
    )
    def post(self, request, *args, **kwargs):
        user = request.user
        token = user.reset_token()
        logger.debug(f"New token for user {user.pk} is {token.key}")
        return Response({"token": token.key}, status=201)


class UserGetTokenAPI(APIView):
    parser_classes = (JSONParser,)
    permission_classes = (IsAuthenticated,)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Get user token",
        operation_description="Get a user token to authenticate to the API as a specific user.",
        responses={
            200: openapi.Response(
                description="User token response",
                schema=openapi.Schema(description="User token", type=openapi.TYPE_STRING),
            )
        },
    )
    def get(self, request, *args, **kwargs):
        user = request.user
        token = Token.objects.get(user=user)
        return Response(str(token), status=200)


class UserWhoAmIAPI(generics.RetrieveAPIView):
    parser_classes = (JSONParser, FormParser, MultiPartParser)
    queryset = User.objects.all()
    permission_classes = (IsAuthenticated,)
    serializer_class = UserSerializer

    def get_object(self):
        return User.objects.get(id=self.request.user.id)

    @swagger_auto_schema(
        tags=["Users"],
        operation_summary="Retrieve my user",
        operation_description="Retrieve details the account that you are using to access the API.",
    )
    def get(self, request, *args, **kwargs):
        return super(UserWhoAmIAPI, self).get(request, *args, **kwargs)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
import pytest
import ujson as json
import requests_mock
import re
import boto3
import logging

from moto import mock_s3
from copy import deepcopy
from django.conf import settings
from projects.models import Project
from tasks.models import Task
from tests.utils import make_project
from users.models import User
from organizations.models import Organization
from types import SimpleNamespace

from .utils import (
    create_business,
    signin,
    gcs_client_mock,
    ml_backend_mock,
    register_ml_backend_mock,
    azure_client_mock,
    redis_client_mock,
)


boto3.set_stream_logger("botocore.credentials", logging.DEBUG)


@pytest.fixture(scope="function")
def aws_credentials():
    """Mocked AWS Credentials for moto."""
    os.environ["AWS_ACCESS_KEY_ID"] = "testing"
    os.environ["AWS_SECRET_ACCESS_KEY"] = "testing"
    os.environ["AWS_SECURITY_TOKEN"] = "testing"
    os.environ["AWS_SESSION_TOKEN"] = "testing"


@pytest.fixture(autouse=True)
def azure_credentials():
    """Mocked Azure credentials"""
    os.environ["AZURE_BLOB_ACCOUNT_NAME"] = "testing"
    os.environ["AZURE_BLOB_ACCOUNT_KEY"] = "testing"


@pytest.fixture(scope="function")
def s3(aws_credentials):
    with mock_s3():
        yield boto3.client("s3", region_name="us-east-1")


@pytest.fixture(autouse=True)
def s3_with_images(s3):
    """
    Bucket structure:
    s3://pytest-s3-images/image1.jpg
    s3://pytest-s3-images/subdir/image1.jpg
    s3://pytest-s3-images/subdir/image2.jpg
    """
    bucket_name = "pytest-s3-images"
    s3.create_bucket(Bucket=bucket_name)
    s3.put_object(Bucket=bucket_name, Key="image1.jpg", Body="123")
    s3.put_object(Bucket=bucket_name, Key="subdir/image1.jpg", Body="456")
    s3.put_object(Bucket=bucket_name, Key="subdir/image2.jpg", Body="789")
    yield s3


@pytest.fixture(autouse=True)
def s3_with_jsons(s3):
    """
    Bucket structure:
    s3://pytest-s3-images/image1.jpg
    s3://pytest-s3-images/subdir/image1.jpg
    s3://pytest-s3-images/subdir/image2.jpg
    """
    bucket_name = "pytest-s3-jsons"
    s3.create_bucket(Bucket=bucket_name)
    s3.put_object(
        Bucket=bucket_name,
        Key="test.json",
        Body=json.dumps({"image_url": "http://ggg.com/image.jpg"}),
    )
    yield s3


@pytest.fixture(autouse=True)
def s3_export_bucket(s3):
    bucket_name = "pytest-export-s3-bucket"
    s3.create_bucket(Bucket=bucket_name)
    yield s3


@pytest.fixture(autouse=True)
def gcs_client():
    with gcs_client_mock():
        yield


@pytest.fixture(autouse=True)
def azure_client():
    with azure_client_mock():
        yield


@pytest.fixture(autouse=True)
def redis_client():
    with redis_client_mock():
        yield


@pytest.fixture(autouse=True)
def ml_backend():
    with ml_backend_mock() as m:
        yield m


@pytest.fixture(autouse=True)
def ml_backend_1(ml_backend):
    register_ml_backend_mock(
        ml_backend,
        url="https://test.heartex.mlbackend.com:9090",
        setup_model_version="Fri Feb 19 17:10:44 2021",
    )
    register_ml_backend_mock(
        ml_backend, url="https://test.heartex.mlbackend.com:9091", health_connect_timeout=True
    )
    register_ml_backend_mock(ml_backend, url="http://localhost:8999", predictions={"results": []})
    yield ml_backend


def pytest_configure():
    for q in settings.RQ_QUEUES.values():
        q["ASYNC"] = False


class URLS:
    """This class keeps urls with api"""

    def __init__(self):
        self.project_create = "/api/projects/"
        self.task_bulk = None

    def set_project(self, pk):
        self.task_bulk = f"/api/projects/{pk}/tasks/bulk/"
        self.plots = f"/projects/{pk}/plots"


def project_ranker():
    label = """<View>
         <HyperText name="hypertext_markup" value="$markup"></HyperText>
         <List name="ranker" value="$replies" elementValue="$text" elementTag="Text" 
               ranked="true" sortedHighlightColor="#fcfff5"></List>
        </View>"""
    return {"label_config": label, "title": "test"}


def project_dialog():
    """Simple project with dialog configs

    :return: config of project with task
    """
    label = """<View>
      <TextEditor>
        <Text name="dialog" value="$dialog"></Text>
        <Header name="header" value="Your answer is:"></Header>
        <TextArea name="answer"></TextArea>
      </TextEditor>
    </View>"""

    return {"label_config": label, "title": "test"}


def project_choices():
    label = """<View>
    <Choices name="animals" toName="xxx" choice="single-radio">
      <Choice value="Cat"></Choice>
      <Choice value="Dog"></Choice>
      <Choice value="Opossum"></Choice>
      <Choice value="Mouse"></Choice>
      <Choice value="Human"/>
    </Choices>

    <Choices name="things" toName="xxx" choice="single-radio">
      <Choice value="Chair"></Choice>
      <Choice value="Car"></Choice>
      <Choice value="Lamp"></Choice>
      <Choice value="Guitar"></Choice>
      <Choice value="None"/>
    </Choices>
    
    <Image name="xxx" value="$image"></Image>
    </View>"""
    return {"label_config": label, "title": "test"}


def setup_project(client, project_template, do_auth=True):
    """Create new test@gmail.com user, login via client, create test project.
    Project configs are thrown over params and automatically grabs from functions names started with 'project_'

    :param client: fixture with http client (from pytest-django package) and simulation of http server
    :param project_template: dict with project config
    :param do_auth: make authorization for creating user
    """
    client = deepcopy(client)
    email = "test@gmail.com"
    password = "test"
    urls = URLS()
    project_config = project_template()

    # we work in empty database, so let's create business user and login
    user = User.objects.create(email=email)
    user.set_password(password)  # set password without hash

    create_business(user)
    org = Organization.create_organization(created_by=user, title=user.first_name)
    user.active_organization = org
    user.save()

    if do_auth:

        assert signin(client, email, password).status_code == 302
        # create project
        with requests_mock.Mocker() as m:
            m.register_uri(
                "POST",
                re.compile(r"ml\.heartex\.net/\d+/validate"),
                text=json.dumps({"status": "ok"}),
            )
            m.register_uri(
                "GET", re.compile(r"ml\.heartex\.net/\d+/health"), text=json.dumps({"status": "UP"})
            )
            r = client.post(urls.project_create, data=project_config)
            print("Project create with status code:", r.status_code)
            assert (
                r.status_code == 201
            ), f"Create project result should be redirect to the next page"

        # get project id and prepare url
        project = Project.objects.filter(title=project_config["title"]).first()
        urls.set_project(project.pk)
        print("Project id:", project.id)

        client.project = project

    client.user = user
    client.urls = urls
    client.project_config = project_config
    client.org = org
    return client


@pytest.fixture
def setup_project_dialog(client):
    return setup_project(client, project_dialog)


@pytest.fixture
def setup_project_for_token(client):
    return setup_project(client, project_dialog, do_auth=False)


@pytest.fixture
def setup_project_ranker(client):
    return setup_project(client, project_ranker)


@pytest.fixture
def setup_project_choices(client):
    return setup_project(client, project_choices)


@pytest.fixture
def business_client(client):
    # we work in empty database, so let's create business user and login
    client = deepcopy(client)
    email = "business@pytest.net"
    password = "pytest"
    user = User.objects.create(email=email)
    user.set_password(password)  # set password without hash
    business = create_business(user)

    user.save()
    org = Organization.create_organization(created_by=user, title=user.first_name)
    client.business = business if business else SimpleNamespace(admin=user)
    client.team = None if business else SimpleNamespace(id=1)
    client.admin = user
    client.annotator = user
    client.user = user
    client.organization = org

    assert signin(client, email, password).status_code == 302
    return client


@pytest.fixture
def annotator_client(client):
    # we work in empty database, so let's create business user and login
    client = deepcopy(client)
    email = "annotator@pytest.net"
    password = "pytest"
    user = User.objects.create(email=email)
    user.set_password(password)  # set password without hash
    user.save()
    business = create_business(user)
    Organization.create_organization(created_by=user, title=user.first_name)
    assert signin(client, email, password).status_code == 302
    client.user = user
    client.annotator = user
    return client


@pytest.fixture
def annotator2_client(client):
    # we work in empty database, so let's create business user and login
    client = deepcopy(client)
    email = "annotator2@pytest.net"
    password = "pytest"
    user = User.objects.create(email=email)
    user.set_password(password)  # set password without hash
    user.save()
    business = create_business(user)
    Organization.create_organization(created_by=user, title=user.first_name)
    assert signin(client, email, password).status_code == 302
    client.user = user
    client.annotator = user
    return client


@pytest.fixture(params=["business", "annotator"])
def any_client(request, business_client, annotator_client):
    if request.param == "business":
        return business_client
    elif request.param == "annotator":
        return annotator_client


@pytest.fixture
def configured_project(business_client, annotator_client):
    _project_for_text_choices_onto_A_B_classes = dict(
        title="Test",
        label_config="""
            <View>
              <Text name="meta_info" value="$meta_info"></Text>
              <Text name="text" value="$text"></Text>
              <Choices name="text_class" toName="text" choice="single">
                <Choice value="class_A"></Choice>
                <Choice value="class_B"></Choice>
              </Choices>
            </View>""",
    )
    _2_tasks_with_textA_and_textB = [
        {"meta_info": "meta info A", "text": "text A"},
        {"meta_info": "meta info B", "text": "text B"},
    ]

    # get user to be owner
    users = User.objects.filter(
        email="business@pytest.net"
    )  # TODO: @nik: how to get proper email for business here?
    project = make_project(_project_for_text_choices_onto_A_B_classes, users[0])

    assert project.ml_backends.first().url == "http://localhost:8999"

    Task.objects.bulk_create(
        [Task(data=task, project=project) for task in _2_tasks_with_textA_and_textB]
    )
    return project


@pytest.fixture(name="django_live_url")
def get_server_url(live_server):
    yield live_server.url



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import json
import time

from uuid import uuid4
from django.http import HttpResponsePermanentRedirect
from django.utils.deprecation import MiddlewareMixin
from django.core.handlers.base import BaseHandler
from django.core.exceptions import MiddlewareNotUsed
from django.middleware.common import CommonMiddleware
from django.conf import settings
from django.utils import timezone
from django.utils.http import escape_leading_slashes
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response
from core.utils.contextlog import ContextLog


class DisableCSRF(MiddlewareMixin):
    def process_request(self, request):
        setattr(request, "_dont_enforce_csrf_checks", True)


class HttpSmartRedirectResponse(HttpResponsePermanentRedirect):
    pass


class CommonMiddlewareAppendSlashWithoutRedirect(CommonMiddleware):
    """This class converts HttpSmartRedirectResponse to the common response
    of Django view, without redirect. This is necessary to match status_codes
    for urls like /url?q=1 and /url/?q=1. If you don't use it, you will have 302
    code always on pages without slash.
    """

    response_redirect_class = HttpSmartRedirectResponse

    def __init__(self, *args, **kwargs):
        # create django request resolver
        self.handler = BaseHandler()

        # prevent recursive includes
        old = settings.MIDDLEWARE
        name = self.__module__ + "." + self.__class__.__name__
        settings.MIDDLEWARE = [i for i in settings.MIDDLEWARE if i != name]

        self.handler.load_middleware()

        settings.MIDDLEWARE = old
        super(CommonMiddlewareAppendSlashWithoutRedirect, self).__init__(*args, **kwargs)

    def get_full_path_with_slash(self, request):
        """Return the full path of the request with a trailing slash appended
        without Exception in Debug mode
        """
        new_path = request.get_full_path(force_append_slash=True)
        # Prevent construction of scheme relative urls.
        new_path = escape_leading_slashes(new_path)
        return new_path

    def process_response(self, request, response):
        response = super(CommonMiddlewareAppendSlashWithoutRedirect, self).process_response(
            request, response
        )

        if isinstance(response, HttpSmartRedirectResponse):
            if not request.path.endswith("/"):
                # remove prefix SCRIPT_NAME
                path = (
                    request.path[len(settings.FORCE_SCRIPT_NAME) :]
                    if settings.FORCE_SCRIPT_NAME
                    else request.path
                )
                request.path = path + "/"
            # we don't need query string in path_info because it's in request.GET already
            request.path_info = request.path
            response = self.handler.get_response(request)

        return response


class SetSessionUIDMiddleware(CommonMiddleware):
    def process_request(self, request):
        if "uid" not in request.session:
            request.session["uid"] = str(uuid4())


class ContextLogMiddleware(CommonMiddleware):
    def __init__(self, get_response):
        self.get_response = get_response
        self.log = ContextLog()

    def __call__(self, request):
        try:
            body = json.loads(request.body)
        except:
            body = {}
        response = self.get_response(request)
        self.log.send(request=request, response=response, body=body)
        return response


class DRFResponseFormatter(CommonMiddleware):
    """This class takes DRF Response and formats it to standard presentation.
    For example, if response = "test string" then response will be reformatted to {"detail": "test string"}
    """

    def __init__(self, get_response):
        super(DRFResponseFormatter, self).__init__(get_response)
        self.get_response = get_response

    def process_response(self, request, response):
        # affect only DRF Response
        if isinstance(response, Response):
            if isinstance(response.data, str):
                response.data = {"detail": response.data}
                response._is_rendered = False
                response.render()

        return response


class DatabaseIsLockedRetryMiddleware(CommonMiddleware):
    """Workaround for sqlite performance issues
    we wait and retry request if database is locked"""

    def __init__(self, get_response):
        if settings.DJANGO_DB != settings.DJANGO_DB_SQLITE:
            raise MiddlewareNotUsed()
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        retries_number = 0
        sleep_time = 1
        backoff = 1.5
        while (
            response.status_code == 500
            and hasattr(response, "content")
            and b"database-is-locked-error" in response.content
            and retries_number < 15
        ):
            time.sleep(sleep_time)
            response = self.get_response(request)
            retries_number += 1
            sleep_time *= backoff
        return response


class UpdateLastActivityMiddleware(CommonMiddleware):
    def process_view(self, request, view_func, view_args, view_kwargs):
        if hasattr(request, "user") and request.method not in SAFE_METHODS:
            if request.user.is_authenticated:
                request.user.update_last_activity()



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

logger = logging.getLogger(__name__)


def get_object_with_check_and_log(request, queryset, *filter_args, **filter_kwargs):
    """Custom get function:
    - if project is deleted - return 404 instead
    - activity log payload is created via 'request_permissions_add'
    """
    from rest_framework.generics import get_object_or_404 as get_object_or_404_rest
    from core.utils.common import get_project, request_permissions_add

    obj = get_object_or_404_rest(queryset, *filter_args, **filter_kwargs)
    # check current Project, raise 404 if it's deleted or create activity log payload
    try:
        project = get_project(obj)
    except AttributeError:
        logger.debug("Can't get Project")
        pass
    else:
        request_permissions_add(request, "project", project)

    return obj



# Generated by Django 3.1.4 on 2021-03-18 14:46

from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ("projects", "0008_auto_20210314_1840"),
        ("tasks", "0005_auto_20210309_1239"),
        ("io_storages", "0002_auto_20210311_0530"),
    ]

    operations = [
        migrations.CreateModel(
            name="LocalFilesMixin",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "path",
                    models.TextField(
                        blank=True, help_text="Local path", null=True, verbose_name="path"
                    ),
                ),
                (
                    "regex_filter",
                    models.TextField(
                        blank=True,
                        help_text="Regex for filtering objects",
                        null=True,
                        verbose_name="regex_filter",
                    ),
                ),
                (
                    "use_blob_urls",
                    models.BooleanField(
                        default=False,
                        help_text="Interpret objects as BLOBs and generate URLs",
                        verbose_name="use_blob_urls",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="LocalFilesExportStorage",
            fields=[
                (
                    "localfilesmixin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="io_storages.localfilesmixin",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="Cloud storage title",
                        max_length=256,
                        null=True,
                        verbose_name="title",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Cloud storage description",
                        null=True,
                        verbose_name="description",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Creation time", verbose_name="created at"
                    ),
                ),
                (
                    "last_sync",
                    models.DateTimeField(
                        blank=True,
                        help_text="Last sync finished time",
                        null=True,
                        verbose_name="last sync",
                    ),
                ),
                (
                    "last_sync_count",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Count of tasks synced last time",
                        null=True,
                        verbose_name="last sync count",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="io_storages_localfilesexportstorages",
                        to="projects.project",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("io_storages.localfilesmixin", models.Model),
        ),
        migrations.CreateModel(
            name="LocalFilesImportStorage",
            fields=[
                (
                    "localfilesmixin_ptr",
                    models.OneToOneField(
                        auto_created=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        parent_link=True,
                        primary_key=True,
                        serialize=False,
                        to="io_storages.localfilesmixin",
                    ),
                ),
                (
                    "title",
                    models.CharField(
                        help_text="Cloud storage title",
                        max_length=256,
                        null=True,
                        verbose_name="title",
                    ),
                ),
                (
                    "description",
                    models.TextField(
                        blank=True,
                        help_text="Cloud storage description",
                        null=True,
                        verbose_name="description",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Creation time", verbose_name="created at"
                    ),
                ),
                (
                    "last_sync",
                    models.DateTimeField(
                        blank=True,
                        help_text="Last sync finished time",
                        null=True,
                        verbose_name="last sync",
                    ),
                ),
                (
                    "last_sync_count",
                    models.PositiveIntegerField(
                        blank=True,
                        help_text="Count of tasks synced last time",
                        null=True,
                        verbose_name="last sync count",
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="io_storages_localfilesimportstorages",
                        to="projects.project",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
            bases=("io_storages.localfilesmixin", models.Model),
        ),
        migrations.CreateModel(
            name="LocalFilesImportStorageLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("key", models.TextField(help_text="External link key", verbose_name="key")),
                (
                    "object_exists",
                    models.BooleanField(
                        default=True,
                        help_text="Whether object under external link still exists",
                        verbose_name="object exists",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Creation time", verbose_name="created at"
                    ),
                ),
                (
                    "task",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="io_storages_localfilesimportstoragelink",
                        to="tasks.task",
                    ),
                ),
                (
                    "storage",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links",
                        to="io_storages.localfilesimportstorage",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="LocalFilesExportStorageLink",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "object_exists",
                    models.BooleanField(
                        default=True,
                        help_text="Whether object under external link still exists",
                        verbose_name="object exists",
                    ),
                ),
                (
                    "created_at",
                    models.DateTimeField(
                        auto_now_add=True, help_text="Creation time", verbose_name="created at"
                    ),
                ),
                (
                    "annotation",
                    models.OneToOneField(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="io_storages_localfilesexportstoragelink",
                        to="tasks.annotation",
                    ),
                ),
                (
                    "storage",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="links",
                        to="io_storages.localfilesexportstorage",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from django.contrib.auth.decorators import login_required
from django.contrib.auth import views as auth_views
from django.shortcuts import render, redirect, reverse
from django.contrib import auth
from django.conf import settings
from django.core.exceptions import PermissionDenied
from rest_framework.authtoken.models import Token

from users import forms
from core.utils.common import load_func
from core.permissions import view_with_auth, IsBusiness
from users.functions import proceed_registration
from organizations.models import Organization
from organizations.forms import OrganizationSignupForm


logger = logging.getLogger()


class FPasswordResetView(auth_views.PasswordResetView):
    from_email = "info@labelstud.io"
    template_name = "password/password_reset_form.html"


class FPasswordResetDoneView(auth_views.PasswordResetDoneView):
    template_name = "password/password_reset_done.html"


class FPasswordResetConfirmView(auth_views.PasswordResetConfirmView):
    template_name = "password/password_reset_confirm.html"


class FPasswordResetCompleteView(auth_views.PasswordResetCompleteView):
    template_name = "password/password_reset_complete.html"


@login_required
def logout(request):
    auth.logout(request)
    if settings.HOSTNAME:
        redirect_url = settings.HOSTNAME
        if not redirect_url.endswith("/"):
            redirect_url += "/"
        return redirect(redirect_url)
    return redirect("/")


def user_signup(request):
    """Sign up page"""
    user = request.user
    next_page = request.GET.get("next")
    token = request.GET.get("token")
    next_page = next_page if next_page else reverse("projects:project-index")
    user_form = forms.UserSignupForm()
    organization_form = OrganizationSignupForm()

    if user.is_authenticated:
        return redirect(next_page)

    # make a new user
    if request.method == "POST":
        organization = Organization.objects.first()
        if settings.DISABLE_SIGNUP_WITHOUT_LINK is True:
            if not (token and organization and token == organization.token):
                raise PermissionDenied()

        user_form = forms.UserSignupForm(request.POST)
        organization_form = OrganizationSignupForm(request.POST)

        if user_form.is_valid():
            redirect_response = proceed_registration(
                request, user_form, organization_form, next_page
            )
            if redirect_response:
                return redirect_response

    return render(
        request,
        "users/user_signup.html",
        {
            "user_form": user_form,
            "organization_form": organization_form,
            "next": next_page,
            "token": token,
        },
    )


def user_login(request):
    """Login page"""
    user = request.user
    next_page = request.GET.get("next")
    next_page = next_page if next_page else reverse("projects:project-index")
    login_form = load_func(settings.USER_LOGIN_FORM)
    form = login_form()

    if user.is_authenticated:
        return redirect(next_page)

    if request.method == "POST":
        form = login_form(request.POST)
        if form.is_valid():
            user = form.cleaned_data["user"]
            auth.login(request, user, backend="django.contrib.auth.backends.ModelBackend")

            # user is organization member
            org_pk = Organization.find_by_user(user).pk
            user.active_organization_id = org_pk
            user.save(update_fields=["active_organization"])
            return redirect(next_page)

    return render(request, "users/user_login.html", {"form": form, "next": next_page})


@view_with_auth(["GET", "POST"], (IsBusiness,))
def user_account(request):
    user = request.user

    if user.active_organization is None and "organization_pk" not in request.session:
        return redirect(reverse("main"))

    form = forms.UserProfileForm(instance=user)
    token = Token.objects.get(user=user)

    if request.method == "POST":
        form = forms.UserProfileForm(request.POST, instance=user)
        if form.is_valid():
            form.save()
            return redirect(reverse("user-account"))

    return render(
        request,
        "users/user_account.html",
        {"settings": settings, "user": user, "user_profile_form": form, "token": token},
    )



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import pytest


@pytest.mark.django_db
def test_active_organization_filled(business_client):
    response = business_client.get("/api/users/")
    response_data = response.json()
    assert response_data[0]["active_organization"] == business_client.organization.id



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 3.1.4 on 2021-02-25 13:48

from django.db import migrations, models
import django.utils.timezone
import users.functions
import users.models

from rest_framework.authtoken.models import Token


def add_tokens(apps, schema_editor):
    User = apps.get_model("users", "User")
    all_users = User.objects.all()

    for user_one in all_users:
        if not hasattr(user_one, "auth_token"):
            Token.objects.create(user=user_one)


class Migration(migrations.Migration):
    replaces = [
        ("users", "0001_initial"),
        ("users", "0002_user_last_activity"),
        ("users", "0003_add_tokens_to_all"),
        ("users", "0004_add_removed_user"),
        ("users", "0005_auto_20200731_1943"),
        ("users", "0006_auto_20201015_1553"),
        ("users", "0007_user_activity_at"),
        ("users", "0008_auto_20210211_1606"),
        ("users", "0009_auto_20210219_1237"),
    ]

    initial = True

    dependencies = [
        ("auth", "0009_alter_user_last_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                (
                    "last_login",
                    models.DateTimeField(blank=True, null=True, verbose_name="last login"),
                ),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                ("username", models.CharField(blank=True, max_length=12, verbose_name="username")),
                (
                    "email",
                    models.EmailField(
                        blank=True, max_length=254, unique=True, verbose_name="email address"
                    ),
                ),
                (
                    "first_name",
                    models.CharField(blank=True, max_length=150, verbose_name="first name"),
                ),
                (
                    "last_name",
                    models.CharField(blank=True, max_length=150, verbose_name="last name"),
                ),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                (
                    "date_joined",
                    models.DateTimeField(
                        default=django.utils.timezone.now, verbose_name="date joined"
                    ),
                ),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.Permission",
                        verbose_name="user permissions",
                    ),
                ),
                (
                    "last_activity",
                    models.DateTimeField(
                        default=django.utils.timezone.now,
                        editable=False,
                        verbose_name="last activity",
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "db_table": "htx_user",
            },
            managers=[
                ("objects", users.models.UserManager()),
            ],
        ),
        migrations.RunPython(add_tokens),
        # migrations.RunPython(add_users),  # TODO: flag:ent
        migrations.AlterField(
            model_name="user",
            name="username",
            field=models.CharField(blank=True, max_length=100, verbose_name="username"),
        ),
        migrations.AlterField(
            model_name="user",
            name="first_name",
            field=models.CharField(blank=True, max_length=256, verbose_name="first name"),
        ),
        migrations.AlterField(
            model_name="user",
            name="last_name",
            field=models.CharField(blank=True, max_length=256, verbose_name="last name"),
        ),
        migrations.AlterField(
            model_name="user",
            name="username",
            field=models.CharField(blank=True, max_length=256, verbose_name="username"),
        ),
        migrations.AddField(
            model_name="user",
            name="activity_at",
            field=models.DateTimeField(auto_now=True, verbose_name="last completion activity"),
        ),
        migrations.AddField(
            model_name="user",
            name="phone",
            field=models.CharField(blank=True, max_length=256, verbose_name="phone"),
        ),
        migrations.AddField(
            model_name="user",
            name="avatar",
            field=models.ImageField(blank=True, upload_to=users.functions.hash_upload),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import drf_yasg.openapi as openapi

from drf_yasg.utils import swagger_auto_schema
from rest_framework import generics, status
from rest_framework.parsers import FormParser, JSONParser, MultiPartParser
from rest_framework.views import APIView
from rest_framework.response import Response

from core.permissions import all_permissions
from core.utils.common import get_object_with_check_and_log
from projects.models import Project
from ml.serializers import MLBackendSerializer
from ml.models import MLBackend
from core.utils.common import bool_from_request

logger = logging.getLogger(__name__)


class MLBackendListAPI(generics.ListCreateAPIView):
    parser_classes = (JSONParser, FormParser, MultiPartParser)
    permission_required = all_permissions.projects_change
    serializer_class = MLBackendSerializer
    swagger_schema = None

    def get_queryset(self):
        project_pk = self.request.query_params.get("project")
        project = get_object_with_check_and_log(self.request, Project, pk=project_pk)
        self.check_object_permissions(self.request, project)
        ml_backends = MLBackend.objects.filter(project_id=project.id)
        for mlb in ml_backends:
            mlb.update_state()
        return ml_backends

    def perform_create(self, serializer):
        ml_backend = serializer.save()
        ml_backend.update_state()


class MLBackendDetailAPI(generics.RetrieveUpdateDestroyAPIView):
    """RUD storage by pk specified in URL"""

    parser_classes = (JSONParser, FormParser, MultiPartParser)
    serializer_class = MLBackendSerializer
    permission_required = all_permissions.projects_change
    queryset = MLBackend.objects.all()
    swagger_schema = None

    def get_object(self):
        ml_backend = super(MLBackendDetailAPI, self).get_object()
        ml_backend.update_state()
        return ml_backend

    def perform_update(self, serializer):
        ml_backend = serializer.save()
        ml_backend.update_state()


class MLBackendTrainAPI(APIView):
    """Train

    After you've activated an ML backend, call this API to start training with the already-labeled tasks.
    """

    permission_required = all_permissions.projects_change

    @swagger_auto_schema(
        request_body=openapi.Schema(
            type=openapi.TYPE_OBJECT,
            properties={
                "use_ground_truth": openapi.Schema(
                    type=openapi.TYPE_BOOLEAN,
                    description="Whether to include ground truth annotations in training",
                )
            },
        ),
        responses={
            200: openapi.Response(
                title="Training OK", description="Training has successfully started."
            ),
            500: openapi.Response(
                description="Training error",
                schema=openapi.Schema(
                    title="Error message",
                    desciption="Error message",
                    type=openapi.TYPE_STRING,
                    example="Server responded with an error.",
                ),
            ),
        },
        tags=["Machine Learning"],
    )
    def post(self, request, *args, **kwargs):
        ml_backend = get_object_with_check_and_log(request, MLBackend, pk=self.kwargs["pk"])
        self.check_object_permissions(self.request, ml_backend)

        ml_backend.train()
        return Response(status=status.HTTP_200_OK)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
""" Actions for tasks and annotations provided by data manager.
    All actions are stored in settings.DATA_MANAGER_ACTIONS dict.
    Data manager uses settings.DATA_MANAGER_ACTIONS to know the list of available actions,
    they are called by entry_points from settings.DATA_MANAGER_ACTIONS dict items.
"""
import os
import logging
import traceback as tb

from importlib import import_module

from django.conf import settings

from data_manager.functions import DataManagerException

logger = logging.getLogger("django")


def check_permissions(params, action):
    """Some actions might have a permissions to perform"""
    if "permissions" in action:
        field = action["permissions"]
        return params[field]
    else:
        return True


def get_all_actions(params):
    """Return dict with registered actions

    :param params: dict with permissions and other flags
    """
    # copy and sort by order key
    actions = list(settings.DATA_MANAGER_ACTIONS.values())
    actions = sorted(actions, key=lambda x: x["order"])
    actions = [
        {key: action[key] for key in action if key != "entry_point"}
        for action in actions
        if not action.get("hidden", False) and check_permissions(params, action)
    ]
    # remove experimental features if they are disabled
    if not params.get("experimental_features", False):
        actions = [action for action in actions if not action.get("experimental", False)]
    return actions


def register_action(entry_point, title, order, **kwargs):
    """Register action in global _action instance,
    action_id will be automatically extracted from entry_point function name
    """
    action_id = entry_point.__name__
    if action_id in settings.DATA_MANAGER_ACTIONS:
        logger.debug('Action with id "' + action_id + '" already exists, rewriting registration')

    settings.DATA_MANAGER_ACTIONS[action_id] = {
        "id": action_id,
        "title": title,
        "order": order,
        "entry_point": entry_point,
        **kwargs,
    }


def register_actions_from_dir(base_module, action_dir):
    """Find all python files nearby this file and try to load 'actions' from them"""
    for path in os.listdir(action_dir):
        if ".py" in path and "__init__" not in path:
            name = path[0 : path.find(".py")]  # get only module name to read *.py and *.pyc
            module_actions = import_module(base_module + "." + name).actions

            for action in module_actions:
                register_action(**action)
                logger.debug("Action registered: " + str(action["entry_point"].__name__))


def perform_action(action_id, project, queryset, **kwargs):
    """Perform action using entry point from actions"""
    if action_id not in settings.DATA_MANAGER_ACTIONS:
        raise DataManagerException("Can't find '" + action_id + "' in registered actions")

    try:
        result = settings.DATA_MANAGER_ACTIONS[action_id]["entry_point"](
            project, queryset, **kwargs
        )
    except Exception as e:
        text = "Error while perform action: " + action_id + "\n" + tb.format_exc()
        logger.error(text)
        raise e

    return result


register_actions_from_dir("data_manager.actions", os.path.dirname(__file__))



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import os
from django.contrib.staticfiles.storage import ManifestStaticFilesStorage
from urllib.parse import unquote, urldefrag, urlsplit, urlunsplit


class SkipMissedManifestStaticFilesStorage(ManifestStaticFilesStorage):
    """We need this class to escape missing files from
    django.contrib.staticfiles.finders.FileSystemFinder:
    this class tries to find js/css/png/jpg/... inside of you js/css/...
    """

    # Disable strict cache manifest checking
    manifest_strict = False

    def hashed_name(self, name, content=None, filename=None):
        # `filename` is the name of file to hash if `content` isn't given.
        # `name` is the base name to construct the new hashed filename from.
        parsed_name = urlsplit(unquote(name))
        clean_name = parsed_name.path.strip()
        filename = (filename and urlsplit(unquote(filename)).path.strip()) or clean_name
        opened = content is None
        if opened:
            if not self.exists(filename):
                return ""
            try:
                content = self.open(filename)
            except IOError:
                # Handle directory paths and fragments
                return name
        try:
            file_hash = self.file_hash(clean_name, content)
        finally:
            if opened:
                content.close()
        path, filename = os.path.split(clean_name)
        root, ext = os.path.splitext(filename)
        if file_hash is not None:
            file_hash = ".%s" % file_hash
        hashed_name = os.path.join(path, "%s%s%s" % (root, file_hash, ext))
        unparsed_name = list(parsed_name)
        unparsed_name[2] = hashed_name
        # Special casing for a @font-face hack, like url(myfont.eot?#iefix")
        # http://www.fontspring.com/blog/the-new-bulletproof-font-face-syntax
        if "?#" in name and not unparsed_name[3]:
            unparsed_name[2] += "?"
        return urlunsplit(unparsed_name)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 2.1.3 on 2020-04-17 22:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Organization",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                ("title", models.CharField(max_length=1000, verbose_name="organization title")),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "created_by",
                    models.OneToOneField(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="organization",
                        to=settings.AUTH_USER_MODEL,
                        verbose_name="created_by",
                    ),
                ),
            ],
            options={
                "db_table": "organization",
            },
        ),
        migrations.CreateModel(
            name="OrganizationMember",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "role",
                    models.CharField(
                        choices=[
                            ("Administrator", "Administrator"),
                            ("Manager", "Manager"),
                            ("Coordinator", "Coordinator"),
                            ("Collaborator", "Collaborator"),
                            ("Non Activated", "Not Activated"),
                            ("Disabled", "Disabled"),
                        ],
                        default="Non Activated",
                        max_length=100,
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, verbose_name="created at")),
                ("updated_at", models.DateTimeField(auto_now=True, verbose_name="updated at")),
                (
                    "organization",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, to="organizations.Organization"
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="om_through",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
        migrations.AddField(
            model_name="organization",
            name="users",
            field=models.ManyToManyField(
                related_name="organizations",
                through="organizations.OrganizationMember",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging
import re

from django.db import models
from django.db.models import Aggregate, Count, Exists, OuterRef, Subquery, Avg, Q, F, Value
from django.contrib.postgres.aggregates import ArrayAgg
from django.db.models.fields.json import KeyTransform
from django.db.models.functions import Coalesce
from django.conf import settings

from data_manager.prepare_params import ConjunctionEnum

logger = logging.getLogger(__name__)

operators = {
    "equal": "",
    "not_equal": "",
    "less": "__lt",
    "greater": "__gt",
    "less_or_equal": "__lte",
    "greater_or_equal": "__gte",
    "in": "",
    "not_in": "",
    "empty": "__isnull",
    "contains": "__icontains",
    "not_contains": "__icontains",
}


def preprocess_field_name(raw_field_name, operator, only_undefined_field=False):
    field_name = raw_field_name.replace("filter:tasks:", "")
    if field_name.startswith("data."):
        if only_undefined_field:
            field_name = f"data__{settings.DATA_UNDEFINED_NAME}"
        else:
            field_name = field_name.replace("data.", "data__")

    # append operator
    field_name = "{}{}".format(field_name, operators.get(operator, ""))

    return field_name


def get_fields_for_annotation(prepare_params):
    """Collecting field names to annotate them

    :param prepare_params: structure with filters and ordering
    :return: list of field names
    """
    from tasks.models import Task

    result = []
    # collect fields from ordering
    if prepare_params.ordering:
        ordering_field_name = prepare_params.ordering[0].replace("tasks:", "").replace("-", "")
        result.append(ordering_field_name)

    # collect fields from filters
    if prepare_params.filters:
        for _filter in prepare_params.filters.items:
            filter_field_name = _filter.filter.replace("filter:tasks:", "")
            result.append(filter_field_name)

    # remove duplicates
    result = set(result)

    # we don't need to annotate regular model fields, so we skip them
    skipped_fields = [field.attname for field in Task._meta.fields]
    skipped_fields.append("id")
    skipped_fields.append("file_upload")
    result = [f for f in result if f not in skipped_fields]
    result = [f for f in result if not f.startswith("data.")]

    return result


def apply_ordering(queryset, ordering):
    if ordering:
        field_name = ordering[0].replace("tasks:", "")
        ascending = False if field_name[0] == "-" else True  # detect direction
        field_name = field_name[1:] if field_name[0] == "-" else field_name  # remove direction

        if "data." in field_name:
            field_name = field_name.replace(".", "__", 1)
            only_undefined_field = (
                queryset.exists() and queryset.first().project.only_undefined_field
            )
            if only_undefined_field:
                field_name = re.sub(
                    "data__\w+", f"data__{settings.DATA_UNDEFINED_NAME}", field_name
                )

            # annotate task with data field for float/int/bool ordering support
            json_field = field_name.replace("data__", "")
            queryset = queryset.annotate(ordering_field=KeyTransform(json_field, "data"))
            f = (
                F("ordering_field").asc(nulls_last=True)
                if ascending
                else F("ordering_field").desc(nulls_last=True)
            )

        else:
            f = (
                F(field_name).asc(nulls_last=True)
                if ascending
                else F(field_name).desc(nulls_last=True)
            )

        queryset = queryset.order_by(f)
    else:
        queryset = queryset.order_by("id")

    return queryset


def cast_value(_filter):
    if _filter.type == "Number":
        _filter.value = float(_filter.value)
    elif _filter.type == "Boolean":
        _filter.value = bool(_filter.value)


def apply_filters(queryset, filters):
    if not filters:
        return queryset

    # convert conjunction to orm statement
    filter_expression = Q()
    if filters.conjunction == ConjunctionEnum.OR:
        conjunction = Q.OR
    else:
        conjunction = Q.AND

    only_undefined_field = queryset.exists() and queryset.first().project.only_undefined_field

    for _filter in filters.items:
        # we can also have annotations filters
        if not _filter.filter.startswith("filter:tasks:"):
            continue

        # django orm loop expression attached to column name
        field_name = preprocess_field_name(_filter.filter, _filter.operator, only_undefined_field)

        # in
        if _filter.operator == "in":
            filter_expression.add(
                Q(
                    **{
                        "{}__gte".format(field_name): _filter.value.min,
                        "{}__lte".format(field_name): _filter.value.max,
                    }
                ),
                conjunction,
            )

        # not in
        elif _filter.operator == "not_in":
            filter_expression.add(
                ~Q(
                    **{
                        "{}__gte".format(field_name): _filter.value.min,
                        "{}__lte".format(field_name): _filter.value.max,
                    }
                ),
                conjunction,
            )

        # empty
        elif _filter.operator == "empty":
            if _filter.value == "True":
                filter_expression.add(Q(**{field_name: True}), conjunction)
            else:
                filter_expression.add(~Q(**{field_name: True}), conjunction)

        # starting from not_
        elif _filter.operator.startswith("not_"):
            cast_value(_filter)
            filter_expression.add(~Q(**{field_name: _filter.value}), conjunction)

        # all others
        else:
            cast_value(_filter)
            filter_expression.add(Q(**{field_name: _filter.value}), conjunction)

    logger.debug(f"Apply filter: {filter_expression}")
    queryset = queryset.filter(filter_expression)
    return queryset


class TaskQuerySet(models.QuerySet):
    def prepared(self, prepare_params=None):
        """Apply filters, ordering and selected items to queryset

        :param prepare_params: prepare params with project, filters, orderings, etc
        :return: ordered and filtered queryset
        """
        queryset = self

        # project filter
        if prepare_params.project is not None:
            queryset = queryset.filter(project=prepare_params.project)

        queryset = apply_filters(queryset, prepare_params.filters)
        queryset = apply_ordering(queryset, prepare_params.ordering)

        if not prepare_params.selectedItems:
            return queryset

        # included selected items
        if prepare_params.selectedItems.all is False and prepare_params.selectedItems.included:
            queryset = queryset.filter(id__in=prepare_params.selectedItems.included)

        # excluded selected items
        elif prepare_params.selectedItems.all is True and prepare_params.selectedItems.excluded:
            queryset = queryset.exclude(id__in=prepare_params.selectedItems.excluded)

        return queryset


class GroupConcat(Aggregate):
    function = "GROUP_CONCAT"
    template = "%(function)s(%(distinct)s%(expressions)s)"

    def __init__(self, expression, distinct=False, **extra):
        super().__init__(
            expression,
            distinct="DISTINCT " if distinct else "",
            output_field=models.CharField(),
            **extra,
        )


def annotate_completed_at(queryset):
    from tasks.models import Annotation

    newest = Annotation.objects.filter(task=OuterRef("pk")).order_by("-created_at")
    return queryset.annotate(completed_at=Subquery(newest.values("created_at")[:1]))


def annotate_cancelled_annotations(queryset):
    from tasks.models import Annotation

    cancelled_annotations = Annotation.objects.filter(was_cancelled=True, task=OuterRef("pk"))
    return queryset.annotate(cancelled_annotations=Exists(cancelled_annotations))


def annotate_annotations_results(queryset):
    if settings.DJANGO_DB == settings.DJANGO_DB_SQLITE:
        return queryset.annotate(
            annotations_results=Coalesce(GroupConcat("annotations__result"), Value(""))
        )
    else:
        return queryset.annotate(annotations_results=ArrayAgg("annotations__result"))


def annotate_predictions_results(queryset):
    if settings.DJANGO_DB == settings.DJANGO_DB_SQLITE:
        return queryset.annotate(
            predictions_results=Coalesce(GroupConcat("predictions__result"), Value(""))
        )
    else:
        return queryset.annotate(predictions_results=ArrayAgg("predictions__result"))


def annotate_annotators(queryset):
    if settings.DJANGO_DB == settings.DJANGO_DB_SQLITE:
        return queryset.annotate(
            annotators=Coalesce(GroupConcat("annotations__completed_by"), Value(None))
        )
    else:
        return queryset.annotate(annotators=ArrayAgg("annotations__completed_by"))


def annotate_predictions_score(queryset):
    return queryset.annotate(predictions_score=Avg("predictions__score"))


def dummy(queryset):
    return queryset


settings.DATA_MANAGER_ANNOTATIONS_MAP = {
    "completed_at": annotate_completed_at,
    "cancelled_annotations": annotate_cancelled_annotations,
    "annotations_results": annotate_annotations_results,
    "predictions_results": annotate_predictions_results,
    "predictions_score": annotate_predictions_score,
    "annotators": annotate_annotators,
    "total_annotations": dummy,
    "total_predictions": dummy,
}


def get_annotations_map():
    return settings.DATA_MANAGER_ANNOTATIONS_MAP


def update_annotation_map(obj):
    settings.DATA_MANAGER_ANNOTATIONS_MAP.update(obj)


class PreparedTaskManager(models.Manager):
    def get_queryset(self, fields_for_evaluation=None):
        queryset = TaskQuerySet(self.model)
        annotations_map = get_annotations_map()

        if fields_for_evaluation is None:
            fields_for_evaluation = []

        # default annotations for calculating total values in pagination output
        queryset = queryset.annotate(
            total_annotations=Count(
                "annotations", distinct=True, filter=Q(annotations__was_cancelled=False)
            ),
            total_predictions=Count("predictions", distinct=True),
        )

        # db annotations applied only if we need them in ordering or filters
        for field in fields_for_evaluation:
            function = annotations_map[field]
            queryset = function(queryset)

        return queryset

    def all(self, prepare_params=None):
        """Make a task queryset with filtering, ordering, annotations

        :param prepare_params: prepare params with filters, orderings, etc
        :return: TaskQuerySet with filtered, ordered, annotated tasks
        """
        if prepare_params is None:
            return self.get_queryset()

        fields_for_annotation = get_fields_for_annotation(prepare_params)
        return self.get_queryset(fields_for_annotation).prepared(prepare_params=prepare_params)


class TaskManager(models.Manager):
    def for_user(self, user):
        return self.filter(project__organization=user.active_organization)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
"""bn URL Configuration

The `urlpatterns` list routes URLs to views. For more information please see:
    https://docs.djangoproject.com/en/2.0/topics/http/urls/
Examples:
Function views
    1. Add an import:  from my_app import views
    2. Add a URL to urlpatterns:  path('', views.home, name='home')
Class-based views
    1. Add an import:  from other_app.views import Home
    2. Add a URL to urlpatterns:  path('', Home.as_view(), name='home')
Including another URLconf
    1. Import the include() function: from django.urls import include, path
    2. Add a URL to urlpatterns:  path('blog/', include('blog.urls'))
"""
from django.apps import apps
from django.conf import settings
from django.conf.urls import include
from django.contrib import admin
from django.urls import path, re_path
from django.views.generic.base import RedirectView
from django.views.static import serve
from drf_yasg import openapi
from drf_yasg.views import get_schema_view
from rest_framework.permissions import AllowAny

from core import views

handler500 = "core.views.custom_500"

schema_view = get_schema_view(
    openapi.Info(
        title="Label Studio API",
        default_version="v2",
        contact=openapi.Contact(url="https://labelstud.io"),
        x_logo={"url": "../../static/icons/logo-black.svg"},
    ),
    public=True,
    permission_classes=(AllowAny,),
)

urlpatterns = [
    re_path(r"^$", views.main, name="main"),
    re_path(
        r"^favicon\.ico$", RedirectView.as_view(url="/static/images/favicon.ico", permanent=True)
    ),
    re_path(
        r"^label-studio-frontend/(?P<path>.*)$",
        serve,
        kwargs={"document_root": settings.EDITOR_ROOT, "show_indexes": True},
    ),
    re_path(
        r"^dm/(?P<path>.*)$",
        serve,
        kwargs={"document_root": settings.DM_ROOT, "show_indexes": True},
    ),
    re_path(
        r"^react-app/(?P<path>.*)$",
        serve,
        kwargs={"document_root": settings.REACT_APP_ROOT, "show_indexes": True},
    ),
    re_path(
        r"^static/fonts/roboto/roboto.css$",
        views.static_file_with_host_resolver(
            "static/fonts/roboto/roboto.css", content_type="text/css"
        ),
    ),
    re_path(
        r"^static/(?P<path>.*)$",
        serve,
        kwargs={"document_root": settings.STATIC_ROOT, "show_indexes": True},
    ),
    re_path(r"^", include("organizations.urls")),
    re_path(r"^", include("projects.urls")),
    re_path(r"^", include("data_import.urls")),
    re_path(r"^", include("data_manager.urls")),
    re_path(r"^", include("data_export.urls")),
    re_path(r"^", include("users.urls")),
    re_path(r"^", include("tasks.urls")),
    re_path(r"^", include("io_storages.urls")),
    re_path(r"^", include("ml.urls")),
    re_path(r"data/local-files/", views.localfiles_data, name="localfiles_data"),
    re_path(r"version/", views.version_page, name="version"),  # html page
    re_path(r"api/version/", views.version_page, name="api-version"),  # json response
    re_path(r"health/", views.health, name="health"),
    re_path(r"metrics/", views.metrics, name="metrics"),
    re_path(r"samples/time-series.csv", views.samples_time_series, name="static_time_series"),
    re_path(
        r"^swagger(?P<format>\.json|\.yaml)$",
        schema_view.without_ui(cache_timeout=0),
        name="schema-json",
    ),
    re_path(
        r"^swagger/$", schema_view.with_ui("swagger", cache_timeout=0), name="schema-swagger-ui"
    ),
    path("docs/api/", schema_view.with_ui("redoc", cache_timeout=0), name="schema-redoc"),
    path(
        "docs/",
        RedirectView.as_view(url="/static/docs/public/guide/introduction.html", permanent=False),
    ),
    path("admin/", admin.site.urls),
    path("django-rq/", include("django_rq.urls")),
    re_path(r"^api-auth/", include("rest_framework.urls", namespace="rest_framework")),
]

if settings.DEBUG:
    try:
        import debug_toolbar

        urlpatterns = [path("__debug__/", include(debug_toolbar.urls))] + urlpatterns
    except ImportError:
        pass



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import logging

from django_filters.rest_framework import DjangoFilterBackend
from django.utils.decorators import method_decorator
from rest_framework import viewsets
from rest_framework.decorators import action
from rest_framework.pagination import PageNumberPagination
from rest_framework.response import Response
from rest_framework.views import APIView
from rest_framework.permissions import IsAuthenticated
from drf_yasg.utils import swagger_auto_schema
from django.db.models import Sum
from ordered_set import OrderedSet

from core.utils.common import get_object_with_check_and_log, int_from_request, bool_from_request
from core.permissions import all_permissions, ViewClassPermission
from core.decorators import permission_required
from projects.models import Project
from projects.serializers import ProjectSerializer
from tasks.models import Task, Annotation

from data_manager.functions import get_all_columns, get_prepared_queryset, evaluate_predictions
from data_manager.models import View
from data_manager.serializers import (
    ViewSerializer,
    DataManagerTaskSerializer,
    SelectedItemsSerializer,
)
from data_manager.actions import get_all_actions, perform_action


logger = logging.getLogger(__name__)


class TaskPagination(PageNumberPagination):
    page_size = 100
    page_size_query_param = "page_size"
    total_annotations = 0
    total_predictions = 0

    def paginate_queryset(self, queryset, request, view=None):
        self.total_annotations = (
            queryset.aggregate(all_annotations=Sum("total_annotations"))["all_annotations"] or 0
        )
        self.total_predictions = (
            queryset.aggregate(all_predictions=Sum("total_predictions"))["all_predictions"] or 0
        )
        return super().paginate_queryset(queryset, request, view)

    def get_paginated_response(self, data):
        return Response(
            {
                "total_annotations": self.total_annotations,
                "total_predictions": self.total_predictions,
                "total": self.page.paginator.count,
                "tasks": data,
            }
        )


@method_decorator(
    name="list",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="List views",
        operation_description="List all views for a specific project.",
    ),
)
@method_decorator(
    name="create",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="Create view",
        operation_description="Create a view for a speicfic project.",
    ),
)
@method_decorator(
    name="retrieve",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="Get view",
        operation_description="Get all views for a specific project.",
    ),
)
@method_decorator(
    name="update",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="Put view",
        operation_description="Overwrite view data with updated filters and other information for a specific project.",
    ),
)
@method_decorator(
    name="partial_update",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="Update view",
        operation_description="Update view data with additional filters and other information for a specific project.",
    ),
)
@method_decorator(
    name="destroy",
    decorator=swagger_auto_schema(
        tags=["Data Manager"],
        operation_summary="Delete view",
        operation_description="Delete a view for a specific project.",
    ),
)
class ViewAPI(viewsets.ModelViewSet):
    queryset = View.objects.all()
    serializer_class = ViewSerializer
    filter_backends = [DjangoFilterBackend]
    my_tags = ["Data Manager"]
    filterset_fields = ["project"]
    task_serializer_class = DataManagerTaskSerializer
    permission_required = ViewClassPermission(
        GET=all_permissions.tasks_view,
        POST=all_permissions.tasks_change,
        PATCH=all_permissions.tasks_change,
        PUT=all_permissions.tasks_change,
        DELETE=all_permissions.tasks_delete,
    )

    def perform_create(self, serializer):
        serializer.save(user=self.request.user)

    @swagger_auto_schema(tags=["Data Manager"])
    @action(detail=False, methods=["delete"])
    def reset(self, _request):
        """
        delete:
        Reset project views

        Reset all views for a specific project.
        """
        queryset = self.filter_queryset(self.get_queryset())
        queryset.all().delete()
        return Response(status=204)

    def get_task_queryset(self, request, view):
        return Task.prepared.all(prepare_params=view.get_prepare_tasks_params())

    @swagger_auto_schema(tags=["Data Manager"], responses={200: task_serializer_class(many=True)})
    @action(detail=True, methods=["get"])
    def tasks(self, request, pk=None):
        """
        get:
        Get task list for view

        Retrieve a list of tasks with pagination for a specific view using filters and ordering.
        """
        view = self.get_object()
        queryset = self.get_task_queryset(request, view)
        context = {
            "proxy": bool_from_request(request.GET, "proxy", True),
            "resolve_uri": True,
            "request": request,
        }
        project = view.project

        # paginated tasks
        self.pagination_class = TaskPagination
        page = self.paginate_queryset(queryset)
        if page is not None:
            # retrieve ML predictions if tasks don't have them
            if project.evaluate_predictions_automatically:
                ids = [task.id for task in page]  # page is a list already
                tasks_for_predictions = Task.objects.filter(id__in=ids, predictions__isnull=True)
                evaluate_predictions(tasks_for_predictions)

            serializer = self.task_serializer_class(page, many=True, context=context)
            return self.get_paginated_response(serializer.data)

        # all tasks
        if project.evaluate_predictions_automatically:
            evaluate_predictions(queryset.filter(predictions__isnull=True))
        serializer = self.task_serializer_class(queryset, many=True, context=context)
        return Response(serializer.data)

    @swagger_auto_schema(tags=["Data Manager"], methods=["get", "post", "delete", "patch"])
    @action(detail=True, url_path="selected-items", methods=["get", "post", "delete", "patch"])
    def selected_items(self, request, pk=None):
        """
        get:
        Get selected items

        Retrieve selected tasks for a specified view.

        post:
        Overwrite selected items

        Overwrite the selected items with new data.

        patch:
        Add selected items

        Add selected items to a specific view.

        delete:
        Delete selected items

        Delete selected items from a specific view.
        """
        view = self.get_object()

        # GET: get selected items from tab
        if request.method == "GET":
            serializer = SelectedItemsSerializer(view.selected_items)
            return Response(serializer.data)

        data = request.data
        serializer = SelectedItemsSerializer(data=data, context={"view": view, "request": request})
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        # POST: set whole
        if request.method == "POST":
            view.selected_items = data
            view.save()
            return Response(serializer.validated_data, status=201)

        selected_items = view.selected_items
        if selected_items is None:
            selected_items = {"all": False, "included": []}

        key = "excluded" if data["all"] else "included"
        left = OrderedSet(selected_items.get(key, []))
        right = OrderedSet(data.get(key, []))

        # PATCH: set particular with union
        if request.method == "PATCH":
            # make union
            result = left | right
            view.selected_items = selected_items
            view.selected_items[key] = list(result)
            view.save(update_fields=["selected_items"])
            return Response(view.selected_items, status=201)

        # DELETE: delete specified items
        if request.method == "DELETE":
            result = left - right
            view.selected_items[key] = list(result)
            view.save(update_fields=["selected_items"])
            return Response(view.selected_items, status=204)


class TaskAPI(APIView):
    permission_required = all_permissions.projects_view

    def get_serializer_class(self):
        return DataManagerTaskSerializer

    @swagger_auto_schema(tags=["Data Manager"])
    def get(self, request, pk):
        """
        get:
        Task by ID

        Retrieve a specific task by ID.
        """
        task = Task.prepared.get(id=pk)
        context = {
            "proxy": bool_from_request(request.GET, "proxy", True),
            "resolve_uri": True,
            "completed_by": "full",
            "request": request,
        }

        # get prediction
        if task.project.evaluate_predictions_automatically and not task.predictions.exists():
            evaluate_predictions([task])

        serializer = self.get_serializer_class()(task, many=False, context=context)
        data = serializer.data
        return Response(data)


class ProjectColumnsAPI(APIView):
    permission_required = all_permissions.projects_view

    @swagger_auto_schema(tags=["Data Manager"])
    def get(self, request):
        """
        get:
        Get data manager columns

        Retrieve the data manager columns available for the tasks in a specific project.
        """
        pk = int_from_request(request.GET, "project", 1)
        project = get_object_with_check_and_log(request, Project, pk=pk)
        self.check_object_permissions(request, project)
        data = get_all_columns(project)
        return Response(data)


class ProjectStateAPI(APIView):
    permission_required = all_permissions.projects_view

    @swagger_auto_schema(tags=["Data Manager"])
    def get(self, request):
        """
        get:
        Project state

        Retrieve the project state for data manager.
        """
        pk = int_from_request(request.GET, "project", 1)  # replace 1 to None, it's for debug only
        project = get_object_with_check_and_log(request, Project.objects.with_counts(), pk=pk)
        self.check_object_permissions(request, project)
        data = ProjectSerializer(project).data
        data.update(
            {
                "can_delete_tasks": True,
                "can_manage_annotations": True,
                "can_manage_tasks": True,
                "source_syncing": False,
                "target_syncing": False,
                "task_count": project.tasks.count(),
                "annotation_count": Annotation.objects.filter(task__project=project).count(),
                "config_has_control_tags": len(project.get_control_tags_from_config()) > 0,
            }
        )
        return Response(data)


class ProjectActionsAPI(APIView):
    permission_required = ViewClassPermission(
        GET=all_permissions.projects_view,
        POST=all_permissions.projects_view,
    )

    @swagger_auto_schema(tags=["Data Manager"])
    def get(self, request):
        """
        get:
        Get actions

        Retrieve all the registered actions with descriptions that data manager can use.
        """
        pk = int_from_request(request.GET, "project", 1)  # replace 1 to None, it's for debug only
        project = get_object_with_check_and_log(request, Project, pk=pk)
        self.check_object_permissions(request, project)

        params = {
            "can_delete_tasks": True,
            "can_manage_annotations": True,
            "experimental_feature": False,
        }

        return Response(get_all_actions(params))

    @swagger_auto_schema(tags=["Data Manager"])
    def post(self, request):
        """
        post:
        Post actions

        Perform an action with the selected items from a specific view.
        """
        pk = int_from_request(request.GET, "project", None)
        project = get_object_with_check_and_log(request, Project, pk=pk)
        self.check_object_permissions(request, project)

        queryset = get_prepared_queryset(request, project)

        # no selected items on tab
        if not queryset.exists():
            response = {"detail": "No selected items for specified view"}
            return Response(response, status=404)

        # wrong action id
        action_id = request.GET.get("id", None)
        if action_id is None:
            response = {"detail": 'No action id "' + str(action_id) + '", use ?id=<action-id>'}
            return Response(response, status=422)

        # perform action and return the result dict
        kwargs = {"request": request}  # pass advanced params to actions
        result = perform_action(action_id, project, queryset, **kwargs)
        code = result.pop("response_code", 200)

        return Response(result, status=code)



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from rest_framework import serializers
from ml.models import MLBackend


class MLBackendSerializer(serializers.ModelSerializer):
    def validate(self, attrs):
        attrs = super(MLBackendSerializer, self).validate(attrs)
        url = attrs["url"]
        if MLBackend.healthcheck_(url).is_error:
            raise serializers.ValidationError(
                "Can't connect to ML backend {url}, health check failed. "
                f"Make sure it is up and your firewall is properly configured. "
                f'<a href="https://labelstud.io/guide/ml.html>Learn more</a>'
                f" about how to set up an ML backend."
            )
        project = attrs["project"]
        setup_response = MLBackend.setup_(url, project)
        if setup_response.is_error:
            raise serializers.ValidationError(
                f"Successfully connected to {url} but it doesn't look like a valid ML backend. "
                f"Reason: {setup_response.error_message}.\n"
                f"Check the ML backend server console logs to check the status. There might be\n"
                f"something wrong with your model or it might be incompatible with the current labeling configuration."
            )
        return attrs

    class Meta:
        model = MLBackend
        fields = "__all__"



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
# Generated by Django 3.1.4 on 2021-03-03 07:29

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    replaces = [
        ("data_manager", "0001_initial"),
        ("data_manager", "0002_auto_20210201_1316"),
        ("data_manager", "0003_filter_index"),
        ("data_manager", "0004_auto_20210204_1231"),
        ("data_manager", "0005_view_user"),
    ]

    initial = True

    dependencies = [
        ("projects", "0060_merge_20210126_1328"),
        ("projects", "0058_remove_projecttemplate_business"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Filter",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "column",
                    models.CharField(
                        help_text="Field name", max_length=1024, verbose_name="column"
                    ),
                ),
                (
                    "type",
                    models.CharField(help_text="Field type", max_length=1024, verbose_name="type"),
                ),
                (
                    "operator",
                    models.CharField(
                        help_text="Filter operator", max_length=1024, verbose_name="operator"
                    ),
                ),
                (
                    "value",
                    models.JSONField(
                        default=dict, help_text="Filter value", null=True, verbose_name="value"
                    ),
                ),
                (
                    "index",
                    models.IntegerField(
                        default=0, help_text="To keep filter order", verbose_name="index"
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="FilterGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "conjunction",
                    models.CharField(
                        help_text="Type of conjunction", max_length=1024, verbose_name="conjunction"
                    ),
                ),
                (
                    "filters",
                    models.ManyToManyField(
                        help_text="Connected filters",
                        related_name="filter_groups",
                        to="data_manager.Filter",
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="View",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True, primary_key=True, serialize=False, verbose_name="ID"
                    ),
                ),
                (
                    "data",
                    models.JSONField(
                        default=dict, help_text="Custom view data", null=True, verbose_name="data"
                    ),
                ),
                (
                    "project",
                    models.ForeignKey(
                        help_text="Project ID",
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="views",
                        to="projects.project",
                    ),
                ),
                (
                    "ordering",
                    models.JSONField(
                        default=dict,
                        help_text="Ordering parameters",
                        null=True,
                        verbose_name="ordering",
                    ),
                ),
                (
                    "filter_group",
                    models.ForeignKey(
                        help_text="Groups of filters",
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        to="data_manager.filtergroup",
                    ),
                ),
                (
                    "selected_items",
                    models.JSONField(
                        default=dict,
                        help_text="Selected items",
                        null=True,
                        verbose_name="selected items",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        help_text="User who made this view",
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="views",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
        ),
    ]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.db.models import signals

from tasks.models import Annotation, Prediction, update_is_labeled_after_removing_annotation
from core.utils.common import temporary_disconnect_signal

from data_manager.functions import evaluate_predictions


def retrieve_tasks_predictions(project, queryset, **kwargs):
    """Retrieve predictions by tasks ids

    :param project: project instance
    :param queryset: filtered tasks db queryset
    """
    evaluate_predictions(queryset)
    return {
        "processed_items": queryset.count(),
        "detail": "Retrieved " + str(queryset.count()) + " predictions",
    }


def delete_tasks(project, queryset, **kwargs):
    """Delete tasks by ids

    :param project: project instance
    :param queryset: filtered tasks db queryset
    """
    count = queryset.count()
    # this signal re-save the task back
    with temporary_disconnect_signal(
        signals.post_delete, update_is_labeled_after_removing_annotation, Annotation
    ):
        queryset.delete()

    # remove all tabs if there are no tasks in project
    reload = False
    if not project.tasks.exists():
        project.views.all().delete()
        reload = True

    return {
        "processed_items": count,
        "reload": reload,
        "detail": "Deleted " + str(count) + " tasks",
    }


def delete_tasks_annotations(project, queryset, **kwargs):
    """Delete all annotations by tasks ids

    :param project: project instance
    :param queryset: filtered tasks db queryset
    """
    task_ids = queryset.values_list("id", flat=True)
    annotations = Annotation.objects.filter(task__id__in=task_ids)
    count = annotations.count()
    annotations.delete()
    return {"processed_items": count, "detail": "Deleted " + str(count) + " annotations"}


def delete_tasks_predictions(project, queryset, **kwargs):
    """Delete all predictions by tasks ids

    :param project: project instance
    :param queryset: filtered tasks db queryset
    """
    task_ids = queryset.values_list("id", flat=True)
    predictions = Prediction.objects.filter(task__id__in=task_ids)
    count = predictions.count()
    predictions.delete()
    return {"processed_items": count, "detail": "Deleted " + str(count) + " predictions"}


actions = [
    {
        "entry_point": retrieve_tasks_predictions,
        "title": "Retrieve predictions",
        "order": 90,
        "permissions": "can_manage_annotations",
        "dialog": {
            "text": "Send the selected tasks to all ML backends connected to the project."
            "This operation migth be abruptly interrupted due to a timeout. "
            "The recommended way to get predictions is to update tasks using the Label Studio API."
            '<a href="https://labelstud.io/guide/ml.html>See more in the documentation</a>.'
            "Please confirm your action.",
            "type": "confirm",
        },
    },
    {
        "entry_point": delete_tasks,
        "title": "Delete tasks",
        "order": 100,
        "permissions": "can_delete_tasks",
        "reload": True,
        "dialog": {
            "text": "You are going to delete the selected tasks. Please confirm your action.",
            "type": "confirm",
        },
    },
    {
        "entry_point": delete_tasks_annotations,
        "title": "Delete annotations",
        "order": 101,
        "permissions": "can_manage_annotations",
        "dialog": {
            "text": "You are going to delete all annotations from the selected tasks. Please confirm your action.",
            "type": "confirm",
        },
    },
    {
        "entry_point": delete_tasks_predictions,
        "title": "Delete predictions",
        "order": 102,
        "permissions": "can_manage_annotations",
        "dialog": {
            "text": "You are going to delete all predictions from the selected tasks. Please confirm your action.",
            "type": "confirm",
        },
    },
]



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
from django.shortcuts import render
from rules.contrib.views import permission_required
from rest_framework.permissions import IsAuthenticated
from core.permissions import view_with_auth, all_permissions


@view_with_auth(["GET"], (IsAuthenticated,))
@permission_required(all_permissions.organizations_change)
def organization_people_list(request):
    return render(
        request,
        "organizations/people_list.html",
        {
            "user": request.user,
        },
    )



"""This file and its contents are licensed under the Apache License 2.0. Please see the included NOTICE for copyright information and LICENSE for a copy of the license.
"""
import io
import copy
import pytest
import zipfile
import ujson as json

from rest_framework.authtoken.models import Token

from tasks.models import Task, Annotation, Prediction
from projects.models import Project


def post_data_as_format(setup, format_type, body, archive, multiply_files):
    # post as data
    if format_type == "json_data":
        return setup.post(setup.urls.task_bulk, data=body, content_type="application/json")

    # post as files
    if format_type == "json_file":
        files = {f"upload_file{i}.json": io.StringIO(body) for i in range(0, multiply_files)}
    elif format_type == "csv_file":
        files = {f"upload_file{i}.csv": io.StringIO(body) for i in range(0, multiply_files)}
    elif format_type == "tsv_file":
        files = {f"upload_file{i}.tsv": io.StringIO(body) for i in range(0, multiply_files)}
    elif format_type == "txt_file":
        files = {f"upload_file{i}.txt": io.StringIO(body) for i in range(0, multiply_files)}
    else:
        raise Exception("Incorrect task data format to post")

    # zip: take files below and zip them
    if "zip" in archive:
        file = io.BytesIO()
        ref = zipfile.ZipFile(file, mode="w", compression=zipfile.ZIP_DEFLATED)
        [ref.writestr(name, body.read()) for name, body in files.items()]

        ref.close()
        file.seek(0, 0)
        files = {"upload_file.zip": file}

        # replicate zip file x2
        if "zip_x2" == archive:
            files.update({"upload_file2.zip": copy.deepcopy(file)})

    return setup.post(setup.urls.task_bulk, files)


@pytest.mark.parametrize("multiply_files", [1, 5])
@pytest.mark.parametrize("format_type", ["json_file", "json_data"])
@pytest.mark.parametrize(
    "tasks, status_code, task_count",
    [
        ([{"data": {"dialog": "some"}}], 201, 1),
        ([{"data": {"dialog": "some"}}] * 10, 201, 10),
        ([{"data": {"another_field": "some", "dialog": "some"}}], 201, 1),
        ([{"data": {"dialog": 123}, "created_at": 123}], 201, 1),
        ([{"data": {"another_field": "some"}}] * 10, 400, 0),
        ([{"data": {}}], 400, 0),
        ([{"data": None}], 400, 0),
        (None, 400, 0),
        ([{"data": "string"}], 400, 0),
        ([{}, {}], 400, 0),
        ([{}], 400, 0),
        ({}, 400, 0),
        ([], 400, 0),
        ([{"dialog": "some"}] * 10, 201, 10),
        ({"dialog": "some"}, 201, 1),
        ([{"dialog": "some", "second_field": 123}] * 10, 201, 10),
        ([{"none": "some", "second_field": 123}] * 10, 400, 0),
    ],
)
@pytest.mark.django_db
def test_json_task_upload(
    setup_project_dialog, format_type, tasks, status_code, task_count, multiply_files
):
    """Upload JSON as file and data with one task to project.
    Decorator pytest.mark.django_db means it will be clean DB setup_project_dialog for this test.
    """
    if format_type == "json_data" and multiply_files > 1:
        pytest.skip("Senseless parameter combination")

    r = post_data_as_format(
        setup_project_dialog, format_type, json.dumps(tasks), "none", multiply_files
    )
    print(f"Create json {format_type} tasks result:", r.content)
    assert r.status_code == status_code, f"Upload tasks failed. Response data: {r.data}"
    assert (
        Task.objects.filter(project=setup_project_dialog.project.id).count()
        == task_count * multiply_files
    )


@pytest.mark.parametrize(
    "tasks, status_code, task_count, annotation_count",
    [
        (
            [{"data": {"dialog": "Test"}, "annotations": [{"result": [{"id": "123"}]}]}] * 10,
            201,
            10,
            10,
        ),
        (
            [
                {
                    "data": {"dialog": "Test"},
                    "annotations": [{"result": [{"id": "123"}], "ground_truth": True}],
                }
            ],
            201,
            1,
            1,
        ),
        ([{"data": {"dialog": "Test"}, "annotations": [{"result": "123"}]}], 400, 0, 0),
        ([{"data": {"dialog": "Test"}, "meta": "test"}] * 10, 400, 0, 0),
        ([{"data": {"dialog": "Test"}, "annotations": "test"}] * 10, 400, 0, 0),
        ([{"data": {"dialog": "Test"}, "annotations": [{"trash": "123"}]}] * 10, 400, 0, 0),
    ],
)
@pytest.mark.django_db
def test_json_task_annotation_and_meta_upload(
    setup_project_dialog, tasks, status_code, task_count, annotation_count
):
    """Upload JSON task with annotation to project"""
    format_type = "json_file"
    multiply_files = 1

    r = post_data_as_format(
        setup_project_dialog, format_type, json.dumps(tasks), "none", multiply_files
    )
    print("Create json tasks with annotations result:", r.content)
    assert r.status_code == status_code, "Upload one task with annotation failed"

    # tasks
    tasks_db = Task.objects.filter(project=setup_project_dialog.project.id)
    assert tasks_db.count() == task_count * multiply_files
    for task in tasks_db:
        assert not task.is_labeled, "Task should not be labeled since annotation is ground_truth"

    # annotations
    annotations = Annotation.objects.filter(task__project=setup_project_dialog.project.id)
    assert annotations.count() == annotation_count * multiply_files
    for i, annotation in enumerate(annotations):
        assert annotation.ground_truth


@pytest.mark.parametrize(
    "tasks, status_code, task_count, prediction_count",
    [
        (
            [
                {
                    "data": {"dialog": "Test"},
                    "predictions": [{"result": [{"id": "123"}], "model_version": "test"}],
                }
            ],
            201,
            1,
            1,
        ),
        ([{"data": {"dialog": "Test"}, "predictions": [{"WRONG_FIELD": "123"}]}], 400, 0, 0),
    ],
)
@pytest.mark.django_db
def test_json_task_predictions(
    setup_project_dialog, tasks, status_code, task_count, prediction_count
):
    """Upload JSON task with predictions to project"""
    r = post_data_as_format(setup_project_dialog, "json_file", json.dumps(tasks), "none", 1)
    assert r.status_code == status_code, "Upload one task with prediction failed"

    # predictions
    predictions = Prediction.objects.filter(task__project=setup_project_dialog.project.id)
    assert predictions.count() == prediction_count
    for i, predictions in enumerate(predictions):
        assert predictions.model_version == "test"


@pytest.mark.parametrize("multiply_files", [1, 5])
@pytest.mark.parametrize("archive", ["none"])
@pytest.mark.parametrize("format_type", ["json_file"])
@pytest.mark.parametrize(
    "tasks, status_code, task_count, annotation_count",
    [
        (
            [
                {
                    "data": {"dialog": "Test"},
                    "annotations": [{"result": [{"id": "123"}]}, {"result": [{"id": "456"}]}],
                }
            ]
            * 10,
            201,
            10,
            20,
        ),
        ([{"data": {"dialog": "Test"}, "annotations": [{"trash": "123"}]}] * 10, 400, 0, 0),
    ],
)
@pytest.mark.django_db
def test_archives(
    setup_project_dialog,
    format_type,
    tasks,
    status_code,
    task_count,
    annotation_count,
    archive,
    multiply_files,
):
    """Upload JSON task with annotation to project"""
    multiplier = (2 if "zip_x2" == archive else 1) * multiply_files

    r = post_data_as_format(
        setup_project_dialog, format_type, json.dumps(tasks), archive, multiply_files
    )
    print("Create json tasks with annotations result:", r.content)
    assert r.status_code == status_code, "Upload one task with annotation failed"

    # tasks
    tasks = Task.objects.filter(project=setup_project_dialog.project.id)
    assert tasks.count() == task_count * multiplier
    for task in tasks:
        assert not task.is_labeled, "Task should not be labeled since annotation is ground_truth"

    # annotations
    annotations = Annotation.objects.filter(task__project=setup_project_dialog.project.id)
    assert annotations.count() == annotation_count * multiplier
    for annotation in annotations:
        assert annotation.ground_truth


@pytest.mark.parametrize("multiply_files", [1, 5])
@pytest.mark.parametrize("archive", ["none"])
@pytest.mark.parametrize("format_type", ["csv_file", "tsv_file"])
@pytest.mark.parametrize(
    "tasks, status_code, task_count",
    [
        ("dialog,second\ndialog 1,second 1\ndialog 2,second 2", 201, 2),
        ("dialog,second,class\ndialog 1, second 2, class 1", 201, 1),
        ("here_is_error_in_column_count,second\ndialog 1, second 1, class 1", 400, 0),
        ("empty_rows\n", 400, 0),
        ("", 400, 0),
    ],
)
@pytest.mark.django_db
def test_csv_tsv_task_upload(
    setup_project_dialog, format_type, tasks, status_code, task_count, archive, multiply_files
):
    """Upload CSV/TSV with one task to project"""
    multiplier = (2 if "zip_x2" == archive else 1) * multiply_files

    tasks = (
        tasks if format_type == "csv_file" else tasks.replace(",", "\t")
    )  # prepare tsv file from csv
    r = post_data_as_format(setup_project_dialog, format_type, tasks, archive, multiply_files)
    print(f"Create {format_type} tasks result:", r.content)

    assert (
        r.status_code == status_code
    ), f"Upload one task {format_type} failed. Response data: {r.data}"
    assert (
        Task.objects.filter(project=setup_project_dialog.project.id).count()
        == task_count * multiplier
    )


@pytest.mark.parametrize("multiply_files", [1, 5])
@pytest.mark.parametrize("format_type", ["txt_file"])
@pytest.mark.parametrize(
    "tasks, status_code, task_count", [("my text 1\nmy text 2\nmy text 3", 201, 3), ("", 400, 0)]
)
@pytest.mark.django_db
def test_txt_task_upload(
    setup_project_dialog, format_type, tasks, status_code, task_count, multiply_files
):
    """Upload CSV/TSV with one task to project"""
    multiplier = multiply_files

    r = post_data_as_format(setup_project_dialog, format_type, tasks, "none", multiply_files)
    print(f"Create {format_type} tasks result:", r.content)

    assert (
        r.status_code == status_code
    ), f"Upload one task {format_type} failed. Response data: {r.data}"
    assert (
        Task.objects.filter(project=setup_project_dialog.project.id).count()
        == task_count * multiplier
    )


@pytest.mark.parametrize(
    "tasks, status_code, task_count, max_duration",
    [
        (
            [{"data": {"dialog": "Test"}, "annotations": [{"result": [{"id": "123"}]}]}] * 1000,
            201,
            1000,
            30,
        )
    ],
)
@pytest.mark.django_db
def test_upload_duration(setup_project_dialog, tasks, status_code, task_count, max_duration):
    """Upload JSON task with annotation to project"""
    r = post_data_as_format(setup_project_dialog, "json_data", json.dumps(tasks), "none", 1)
    print("Create json tasks with annotations result:", r.content)
    assert r.status_code == status_code, ("Upload one task with annotation failed", r.content)

    # tasks
    tasks = Task.objects.filter(project=setup_project_dialog.project.id)
    assert tasks.count() == task_count
    for task in tasks:
        assert not task.is_labeled, "Task should not be labeled since annotation is ground_truth"

    # check max duration
    result = json.loads(r.content)
    assert result["duration"] < max_duration, "Max duration of adding tasks is exceeded"


@pytest.mark.parametrize(
    "tasks, status_code, task_count",
    [
        (
            [{"data": {"dialog": "Test"}, "annotations": [{"result": [{"id": "123"}]}]}] * 100,
            201,
            100,
        )
    ],
)
@pytest.mark.django_db
def test_url_upload(mocker, setup_project_dialog, tasks, status_code, task_count):
    """Upload tasks from URL"""

    def info():
        class Info:
            @staticmethod
            def get(name):
                return 12345

        return Info()

    urlopen_return = io.StringIO(json.dumps(tasks))
    urlopen_return.info = info

    with mocker.patch("data_import.uploader.urlopen", return_value=urlopen_return) as m:
        # m.return_value.info.get = mocker.PropertyMock(return_value=lambda x: 12345)

        url = "http://localhost:8111/test.json"
        r = setup_project_dialog.post(
            setup_project_dialog.urls.task_bulk,
            data="url=" + url,
            content_type="application/x-www-form-urlencoded",
        )
        assert r.status_code == status_code, "Upload URL failed: " + str(r.content)

        # tasks
        tasks = Task.objects.filter(project=setup_project_dialog.project.id)
        assert tasks.count() == task_count
        for task in tasks:
            assert (
                not task.is_labeled
            ), "Task should not be labeled since annotation is ground_truth"


@pytest.mark.parametrize(
    "tasks, status_code, task_count, bad_token",
    [([{"dialog": "Test"}] * 1, 201, 1, False), ([{"dialog": "Test"}] * 1, 401, 0, True)],
)
@pytest.mark.django_db
def test_upload_with_token(setup_project_for_token, tasks, status_code, task_count, bad_token):
    """Upload with Django Token"""
    setup = setup_project_for_token
    token = Token.objects.get(user=setup.user)
    token = "Token " + str(token)
    broken_token = "Token broken"
    data = setup.project_config
    data["organization_pk"] = setup.org.pk
    r = setup.post(setup.urls.project_create, data=data, HTTP_AUTHORIZATION=token)
    print("Project create with status code:", r.status_code, r.content)
    assert (
        r.status_code == 201
    ), "Create project result should be redirect to the next page: " + str(r.content)

    project = Project.objects.filter(title=setup.project_config["title"]).first()
    setup.urls.set_project(project.pk)

    r = setup.post(
        setup.urls.task_bulk,
        data=json.dumps(tasks),
        content_type="application/json",
        HTTP_AUTHORIZATION=broken_token if bad_token else token,
    )
    assert r.status_code == status_code, "Create json tasks result: " + str(r.content)

    # tasks
    tasks = Task.objects.filter(project=project.id)
    assert tasks.count() == task_count
